from .formation_plots import *
import logging
import re
import io
import copy
from scipy.stats import zscore
import numpy as np
import pandas as pd
from sklearn.decomposition import PCA
import matplotlib
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.dates as mdates
import xlsxwriter
from bmxp.gravity import spearman, pearson
from bmxp import IMDATA

matplotlib.use("agg")

logging.basicConfig()
LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)

__version__ = "0.2.11"


def report(
    sample_data,
    smdata,
    fmdata,
    dataset_name,
    sample_names=None,
    out_filepath=None,
    write_pdf=True,
):
    """
    Handles /formation endpoint
    """
    fmdata = fmdata.copy()
    smdata = smdata.copy()
    sample_data = sample_data.copy()

    # raise ValueError("'Compound_ID' was not found in your dataset.")
    if "Compound_ID" in fmdata:
        fmdata.set_index("Compound_ID", drop=True, inplace=True)
        sample_data.index = fmdata.index

    if sample_names is None:
        if "reporting_name" in smdata:
            sample_names = smdata["reporting_name"].copy()
        elif "program_id" in smdata:
            sample_names = smdata["program_id"].copy()
        else:
            sample_names = pd.Series(smdata.index, index=smdata.index.copy())
    # check and create necessary columns

    fill_values = {
        "Annotation_ID": "",
        "HMDB_ID": "",
        "Metabolite": "",
    }
    for fill_key in fill_values:
        if fill_key not in fmdata.columns:
            fmdata[fill_key] = fill_values[fill_key]

    fmdata.columns.name = None

    if "raw_file_name" not in smdata.columns:
        smdata["raw_file_name"] = smdata.index
        # raise ValueError("'raw_file_name' was not found in the sample metadata.")

    fill_values = {
        "sample_type": "unknown_type",
        "injection_type": "unknown_type",
        "injection_order": list(range(len(smdata))),
        "column_number": 1,
    }
    for fill_key in fill_values:
        if fill_key not in smdata.columns:
            smdata[fill_key] = fill_values[fill_key]

    # fill necessary missing values
    for sample_type in ("injection_type", "sample_type"):
        smdata[sample_type] = smdata[sample_type].fillna("")

    smdata.set_index("raw_file_name", drop=True, inplace=True)
    smdata = smdata.replace({"": float("nan"), "NA": float("nan")})
    smdata = smdata.convert_dtypes()
    # date_extracted usually does not convert to datetime correctly
    for date_type in ("date_extracted", "date_injected"):
        if date_type not in smdata.columns:
            continue
        try:
            smdata[date_type] = pd.to_datetime(smdata[date_type])
        except Exception as e:  # pylint: disable=broad-except
            raise ValueError(f"{date_type} contains one or more invalid dates.") from e

    try:
        smdata["injection_order"] = pd.to_numeric(smdata["injection_order"])
    except ValueError as e:
        character = re.search('".+"', str(e))
        if character:
            raise ValueError(
                f"Injection_order contains a non-numeric character: "
                f"{character.group()}."
            ) from e
        raise ValueError("Injection_order contains a non-numeric character.") from e

    transposed_s_data = sample_data.T
    transposed_s_data.index.name = "rawfile"
    transposed_s_data = transposed_s_data.apply(lambda x: x.fillna(x.min() / 2))
    # drop empty columns
    transposed_s_data = transposed_s_data.dropna(how="all", axis=1)
    s_data_zscores = zscore(transposed_s_data.to_numpy())
    s_data_zscores = np.nan_to_num(s_data_zscores)  # zscore is NaN if stdev is 0
    sample_pca = PCA(n_components=2)
    sample_pca_data = sample_pca.fit_transform(s_data_zscores)

    pca_df = pd.DataFrame(data=sample_pca_data, columns=["pc1", "pc2"])

    is_indices = fmdata.loc[fmdata["HMDB_ID"].str.lower() == "internal standard"].index
    internal_standard_s_data = sample_data.loc[is_indices]
    file_handle = io.BytesIO()
    palette = [
        # basel, from the R library 'yarrr'
        # https://cran.r-project.org/web/packages/yarrr/vignettes/piratepal.html
        (12, 91, 176, 0.7),  # blue
        (238, 0, 17, 0.7),  # red
        (21, 152, 61, 0.7),  # green
        (236, 87, 154, 0.7),  # pink
        (250, 107, 9, 0.7),  # orange
        (20, 155, 237, 0.7),  # light blue
        (161, 199, 32, 0.7),  # light green
        (254, 183, 11, 0.7),  # yellow
        (22, 160, 140, 0.7),  # teal
        (154, 112, 62, 0.7),  # brown
    ]
    nan_color = (0.55, 0.55, 0.55, 0.3)
    with PdfPages(file_handle) as pdf:
        plt.figure(figsize=(20, 12))
        plt.axis("off")
        plt.text(
            0.5,
            1,
            f"QC Report for {dataset_name}",
            ha="center",
            va="top",
            fontsize=24,
        )
        plt.text(
            0.5,
            0.9,
            f"Number of samples: {len(sample_data.columns)}",
            ha="center",
            va="top",
            fontsize=18,
        )
        plt.text(
            0.5,
            0.8,
            f"Number of features: {len(fmdata.index)}",
            ha="center",
            va="top",
            fontsize=18,
        )
        if "date_injected" in smdata.columns:
            try:
                date = str(smdata["date_injected"].min().strftime("%B %d, %Y"))
            except:
                date = "Unknown"
            plt.text(
                0.5,
                0.70,
                ("Date of first injection: " + date),
                ha="center",
                va="top",
                fontsize=18,
            )
        pdf.savefig()
        plt.close("all")
        # create zoomable PCA plot
        plot_formation_zoomable_plot(
            pca_df,
            "PCA of metabolites labeled by sample name",
            f"PC1 ({round(sample_pca.explained_variance_ratio_[0] * 100)}"
            "% variance explained)",
            f"PC2 ({round(sample_pca.explained_variance_ratio_[1] * 100)}"
            "% variance explained)",
            sample_names,
            pdf,
        )

        components_df = pd.DataFrame(
            data=sample_pca.components_.transpose(),
            columns=["pc1", "pc2"],
            index=transposed_s_data.columns,
        )
        # create zoomable loadings plots
        plot_formation_zoomable_plot(
            components_df,
            "Loadings plot labeled by Annotation_ID",
            "Principal Component 1",
            "Principal Component 2",
            fmdata.loc[transposed_s_data.columns, "Metabolite"],
            pdf,
            rasterized=True,
        )
        plot_formation_zoomable_plot(
            components_df,
            "Loadings plot labeled by Compound_ID",
            "Principal Component 1",
            "Principal Component 2",
            transposed_s_data.columns,
            pdf,
            rasterized=True,
        )

        # create PCAs colored by inj metadata
        num_columns = smdata.select_dtypes(include=["float64", "int64"]).columns
        date_columns = smdata.select_dtypes(include=["datetime"]).columns
        str_columns = smdata.select_dtypes(include=["string"]).columns
        cmap = copy.copy(plt.cm.get_cmap("magma"))
        cmap.set_bad(color=nan_color)
        graph_index = 0
        for i, column_name in enumerate(smdata.columns):
            if i % 6 == 0 and i != 0:
                plt.tight_layout()
                pdf.savefig()
                plt.close("all")
            if i % 6 == 0:
                plt.figure(figsize=(20, 12))

            plt.subplot(2, 3, i % 6 + 1)
            plt.xlabel(
                f"PC1 ({round(sample_pca.explained_variance_ratio_[0] * 100)}"
                "% variance explained)"
            )
            plt.ylabel(
                f"PC2 ({round(sample_pca.explained_variance_ratio_[1] * 100)}"
                "% variance explained)"
            )
            plt.title("PCA of metabolites - colored by " + str(column_name))

            if column_name in num_columns:
                col = smdata[column_name].astype("float").replace({pd.NA: np.nan})
                plt.scatter(
                    pca_df["pc1"],
                    pca_df["pc2"],
                    c=col,
                    s=50,
                    cmap=cmap,
                    plotnonfinite=True,
                    alpha=0.7,
                )
                if col.isnull().values.any():
                    plt.legend(["NA"])
                    axes = plt.gca()
                    leg = axes.get_legend()
                    leg.legend_handles[0].set_color(nan_color)
                plt.colorbar(label=column_name)

            elif column_name in str_columns:
                col = smdata[column_name].replace({pd.NA: "NA"})
                targets = col.unique()
                colors = {
                    targets[k]: tuple(x / 255 for x in palette[k % len(palette)][:-1])
                    + (0.7,)
                    for k in range(len(targets))
                }
                plt.scatter(
                    pca_df["pc1"],
                    pca_df["pc2"],
                    color=col.map(colors),
                    s=50,
                )
                legend_colors = {k: colors[k] for k in list(colors)[:20]}
                handles = [
                    Line2D(
                        [0],
                        [0],
                        marker="o",
                        color="w",
                        markerfacecolor=v,
                        label=k,
                        markersize=8,
                    )
                    for k, v in legend_colors.items()
                ]
                leg = plt.legend(title="color", handles=handles)
                plt.gca().add_artist(leg)

            elif column_name in date_columns:
                plt.scatter(
                    pca_df["pc1"],
                    pca_df["pc2"],
                    c=mdates.date2num(smdata[column_name]),
                    s=50,
                    cmap=cmap,
                    alpha=0.7,
                )
                colorbar = plt.colorbar(label=column_name)
                loc = mdates.AutoDateLocator()
                colorbar.ax.yaxis.set_major_locator(loc)
                colorbar.ax.yaxis.set_major_formatter(mdates.ConciseDateFormatter(loc))
            graph_index = i
        # reset graph_index to create page break between PCAs and other plots
        graph_index = 0
        is_ids = fmdata.loc[internal_standard_s_data.index, "Annotation_ID"]
        internal_standard_s_data.reset_index().apply(
            lambda row: plot_formation_line_plot(
                row,
                graph_index + 3 * row.name,
                smdata,
                palette,
                pdf,
                "injection_type",
                ann_id=is_ids.iloc[row.name],
            ),
            axis=1,
        )
        graph_index += 3 * len(internal_standard_s_data.index)
        norm = internal_standard_s_data.apply(lambda row: row / row.median(), axis=1)
        line_colors = []
        num_inj_types = len(smdata["injection_type"].unique())
        for i in range(num_inj_types, num_inj_types + len(norm.index)):
            line_colors.append(
                tuple(x / 255 for x in palette[i % len(palette)][:-1]) + (0.7,)
            )

        norm.reset_index().apply(
            lambda row: plot_formation_line_plot(
                row,
                graph_index,
                smdata,
                palette,
                pdf,
                "injection_type",
                line_colors[row.name],
                True,
            ),
            axis=1,
        )
        if not internal_standard_s_data.empty:
            plt.ylim(bottom=0)
            leg_labels = [f"{k} - {v}" for k, v in zip(norm.index, is_ids)]
            handles_two = [
                Line2D([0], [0], color=v, label=k)
                for k, v in zip(leg_labels, line_colors)
            ]
            leg2 = plt.legend(title="color", handles=handles_two, loc=4)
            plt.gca().add_artist(leg2)
        graph_index += 3
        sample_medians = sample_data.fillna(0).median()
        plot_formation_line_plot(
            sample_medians,
            graph_index,
            smdata,
            palette,
            pdf,
            "sample_type",
            sample_names=sample_names,
        )
        graph_index += 3
        median_data = sample_data.apply(lambda col: col.fillna(col.min() / 2))
        median_data = median_data.apply(np.log)
        median_data = median_data.T
        unique_samp_types = list(smdata["sample_type"].unique())
        sample_colors = {
            key: tuple(x / 255 for x in value[:-1]) + (0.7,)
            for key, value in zip(unique_samp_types, palette)
        }
        colors = smdata["sample_type"].copy().astype(object).map(sample_colors)
        if len(median_data.index) <= 1500:
            for i in range(0, len(median_data.index), 150):
                plot_formation_quartile(
                    median_data.iloc[i : i + 150].T,
                    graph_index,
                    sample_colors,
                    colors[i : i + 150],
                    pdf,
                    sample_names[i : i + 150],
                )
                graph_index += 3
        plt.tight_layout()
        pdf.savefig()  # saves the current figure into a pdf page
        plt.close("all")
        inj_types = smdata["injection_type"].unique()
        pools = [
            inj_type.upper() for inj_type in inj_types if inj_type.startswith("pref")
        ]
        create_pools_cv_table(fmdata, pools, pdf)
        plt.tight_layout()
        pdf.savefig()  # saves the current figure into a pdf page
        plt.close("all")
    if out_filepath is None:
        out_filepath = f"{dataset_name}_formation_report.pdf"
    if write_pdf:
        with open(out_filepath, "wb") as f:
            f.write(file_handle.getbuffer())
    return file_handle


def report_from_formatted(
    dataset, dataset_name=None, out_filepath=None, write_pdf=True
):
    if isinstance(dataset, pd.DataFrame):
        if dataset_name is None:
            dataset_name = "Dataset"
        df = dataset
    else:
        if dataset_name is None:
            dataset_name = dataset.split("\\")[-1].split("/")[-1]
        try:
            df = pd.read_csv(dataset, header=None)
        except Exception as e:  # pylint: disable=broad-except
            try:
                df = pd.read_excel(dataset, engine="openpyxl", header=None)
            except:
                raise ValueError(
                    "Your file does not appear to be a csv or xlsx."
                ) from e

    # Find the index of the first non-missing value
    row = df.iloc[0, :].copy()
    row.replace("", np.nan, inplace=True)
    pivot_col = df.columns.get_loc(row.first_valid_index())

    row = df.iloc[:, 0].copy()
    row.replace("", np.nan, inplace=True)
    pivot_row = df.index.get_loc(row.first_valid_index())

    pivot_value = df.iloc[pivot_row, pivot_col].lower()

    if pivot_col < 0 or pivot_row < 0:
        raise ValueError("Your dataset is blank on either the first row or column.")

    feature_metadata = df.iloc[pivot_row + 1 :, : pivot_col + 1].copy()
    feature_metadata.columns = df.iloc[pivot_row, : pivot_col + 1]

    inj_metadata = df.iloc[: pivot_row + 1, pivot_col:].copy()
    inj_metadata = inj_metadata.transpose().reset_index(drop=True)
    inj_metadata.columns = inj_metadata.iloc[0]
    inj_metadata = inj_metadata.drop(0)
    inj_metadata.columns = inj_metadata.columns.str.lower()
    # makes column name unique - https://stackoverflow.com/questions/24685012
    cols = pd.Series(inj_metadata.columns)
    for dup in inj_metadata.columns[inj_metadata.columns.duplicated(keep=False)]:
        cols[inj_metadata.columns.get_loc(dup)] = [
            dup + "." + str(d_idx) if d_idx != 0 else dup
            for d_idx in range(inj_metadata.columns.get_loc(dup).sum())
        ]
    inj_metadata.columns = cols
    # check and create necessary columns

    sample_data = df.iloc[pivot_row + 1 :, pivot_col + 1 :]
    sample_names = df.iloc[pivot_row, pivot_col + 1 :].astype(str)
    sample_data.columns = inj_metadata.index.astype(str) + " (" + sample_names + ")"
    sample_data.index = feature_metadata.index.copy()
    sample_data = sample_data.replace("", "nan")
    sample_data = sample_data.astype(float)
    inj_metadata = inj_metadata.drop(pivot_value, axis=1)

    return report(
        sample_data,
        inj_metadata,
        feature_metadata,
        dataset_name,
        sample_names,
        out_filepath,
        write_pdf,
    )


def _sort_dataset(final_dataset):
    """Sort annotated metabolites by order, then, class MZ, RT;
    sort nontargeted by RT, MZ"""
    # force internal standards sort to top
    final_dataset.loc[
        final_dataset["HMDB_ID"].str.lower() == "internal standard", "superClass"
    ] = "AAAAAAAAAAAAAA"

    # make sure all annotated features are above nontargeted features
    final_dataset = final_dataset.sort_values(by=["Metabolite"])

    annotated = ~pd.isnull(final_dataset["Metabolite"])
    # sort annotated if present
    try:
        final_dataset.loc[annotated, :] = (
            final_dataset.loc[annotated, :]
            .sort_values(
                by=[
                    "orderNum",
                    "superClass",
                    "mainClass",
                    "subClass",
                    "MZ_Calculated",
                    "RT",
                ],
                key=lambda col: (
                    col
                    if not pd.api.types.is_string_dtype(col.fillna(""))
                    else col.str.lower()
                ),
            )
            .values
        )
    except KeyError:
        pass

    # sort nontargeted separately
    final_dataset.loc[~annotated, :] = (
        final_dataset.loc[~annotated, :].sort_values(by=["RT", "MZ"]).values
    )

    return final_dataset


def harmonize_metadata(data, injectionset, sampleset):
    """Generates combined Sampleset and Injectionset metadata with columns matching
    data. Also removes not_used. If there are samples in data not present in either
    sampleset or injectionset, they are preserved and missing metadata is null.
    """
    warnings = []
    # rename column for older datasets
    if "reporting_name" not in injectionset.columns:
        injectionset = injectionset.rename(columns={"program_id": "reporting_name"})
    combined = injectionset.loc[data.columns].copy()
    combined["broad_id"] = combined["broad_id"].fillna("NA")

    s_broad_ids = combined.loc[combined["injection_type"] == "sample", "broad_id"]
    duplicates = set(s_broad_ids.loc[s_broad_ids.duplicated()]) - set([""])
    if len(duplicates) > 0:
        warnings.append(f"There are duplicate Broad_IDs: {', '.join(duplicates)}")

    sampleset_copy = sampleset.copy()
    sampleset_copy.loc["", :] = np.nan  # blank broad_id means no metadata
    combined = combined.join(sampleset_copy, on="broad_id", lsuffix="_inj")

    additional_meta = [
        col for col in sampleset.columns if col not in ["program_id", "broad_id"]
    ]

    # use injection set reporting_name for pools or missing program_ids if present
    combined["program_id"] = combined["program_id"].astype(object)
    pools = combined.index[combined["broad_id"].str.lower() == "pref"]
    missing = pd.isnull(combined["program_id"])
    if "reporting_name" in combined.columns:
        combined.loc[missing, "program_id"] = combined.loc[missing, "reporting_name"]
    # name pools if possible
    elif pools.str.contains("PREF").all():
        combined.loc[pools, "program_id"] = pools.str.extract(r"(PREF\w*)").values
        warnings.append(
            "PREF program_id names were determined from injection_id names. Please "
            + "review and confirm program_id names before sharing results."
        )

    if pd.isnull(combined.loc[:, "program_id"]).any():
        warnings.append(
            "Not all program_ids could be filled. Please add any missing program_ids "
            "before sharing results."
        )
    combined[additional_meta] = combined[additional_meta].astype(object).fillna("NA")

    qcrole = IMDATA["QCRole"]
    qcrole_map = {
        "tube_blank": "Tube_Blank",
        "blank": "Blank",
        "not_used": "Not_Used",
        "mm": "Master_Mix",
        "ms2": "MS2",
        "brpp": "BRPP",
    }

    if qcrole in combined.columns:
        for prefix, replacement in qcrole_map.items():
            mask = combined["injection_type"].str.startswith(prefix, na=False)
            combined.loc[mask, qcrole] = replacement
    else:
        combined[qcrole] = combined["injection_type"]

    # re-order
    injection_meta = combined.reindex(
        columns=[
            "date_extracted",
            "date_injected",
            "column_number",
            "injection_order",
            "injection_type",
            qcrole,
            "broad_id",
        ]
        + additional_meta
        + [
            "raw_file_name",
            "program_id",
        ]
    )
    injection_meta["raw_file_name"] = injection_meta.index
    # rename sample metadata and QCRole columns
    names = {col: col.lower().replace(" ", "_") for col in additional_meta}
    names[qcrole] = "sample_type"
    injection_meta = injection_meta.rename(columns=names)

    injection_meta.loc[:, "date_extracted"] = pd.to_datetime(
        injection_meta["date_extracted"],
        errors="coerce",
    )
    injection_meta.loc[:, "date_injected"] = pd.to_datetime(
        injection_meta["date_injected"],
        errors="coerce",
    )

    return injection_meta, warnings


def fill_f_mdata(metadata, annotations):
    """combine feature metadata from dataset and database queries for final results"""
    # request annotations and their HMBD IDs
    main_cols = ["__annotation_id", "Compound_ID", "MZ", "RT"]
    additional_meta = [col for col in metadata.columns if col not in main_cols]
    feature_meta = metadata.loc[:, main_cols + additional_meta]
    feature_meta["superClass"] = None
    feature_meta["mainClass"] = None
    feature_meta["subClass"] = None
    feature_meta["Adduct_Priority"] = None
    feature_meta["HMDB_ID"] = None
    feature_meta["HMDB_specificity (1=match; 2=representative)"] = None
    feature_meta["Annotation_ID"] = None
    feature_meta["Metabolite"] = None

    # move method to front
    method_col = ""
    if "Method" in feature_meta:
        method_col = feature_meta.pop("Method")
    feature_meta.insert(0, "Method", method_col)

    for key, value in annotations.items():
        if key not in feature_meta["__annotation_id"].unique():
            continue
        to_fill = feature_meta["__annotation_id"] == key
        for anno_key in value:
            feature_meta.loc[to_fill, anno_key] = value[anno_key]

    # format labels for each column; None = default formatting
    # feature_meta = feature_meta.reset_index(drop=True)
    return feature_meta


def label_most_abundant(data, metadata, corr_method="spearman"):
    """Creates Primary and Corr_To_Primary columns given a clustered dataset. "Primary"
    feature is annotated or has the highest mean abundance.

    :param data: DataFrame, data columns that were used for clustering
    :param metadata: DataFrame, contains columns "Annotation_ID", "Cluster_Num", and
    "Cluster_Size" at minimum
    :param corr_method: str, correlation method used in clustering, "spearman" or
    "pearson", defaults to "spearman"
    :return: DataFrame, Primary and Corr_To_Primary columns
    """
    metadata = metadata.copy()
    # explicitly cast as str in case the Annotation_ID column is empty
    metadata["Annotation_ID"] = metadata["Annotation_ID"].fillna("")
    # identify the primary adduct
    for cluster_num in metadata["Cluster_Num"].unique():
        if not cluster_num and cluster_num != 0:
            continue
        cluster_index = metadata.index[metadata["Cluster_Num"] == cluster_num]
        key = (metadata["Cluster_Num"] == cluster_num).values & (
            metadata["Annotation_ID"] > ""
        ).values

        if key.any():
            metadata.loc[key, "Primary"] = True
            max_idx = metadata.index[key][0]
        else:
            max_idx = data.loc[cluster_index, :].fillna(0).mean(axis="columns").idxmax()
            metadata.loc[max_idx, "Primary"] = True

        for feature in cluster_index:
            arr1 = data.loc[feature, :].values.astype(np.float64)
            arr2 = data.loc[max_idx, :].values.astype(np.float64)
            if corr_method == "spearman":
                metadata.loc[feature, "Corr_To_Primary"] = spearman(
                    arr1, arr2, "drop", legacy_mode=True
                )
            else:
                metadata.loc[feature, "Corr_To_Primary"] = pearson(arr1, arr2)

    # label the singletons
    metadata.loc[metadata["Cluster_Size"] == 1, "Primary"] = True

    return metadata[["Primary", "Corr_To_Primary"]]


def apply_pool_instructions(data, feature_instructions, inplace=False):
    """Apply all pool instructions"""
    if feature_instructions is None:
        return data
    if not inplace:
        data = data.copy()
    for feature, instructions in feature_instructions.items():
        if feature not in data.index:
            continue
        for inst in instructions:
            if inst["op1"] not in data.columns or (
                inst["instruction"] != "delete" and inst["op2"] not in data.columns
            ):
                # we're probably doing final formatting and already removed some unused
                # pools, so just move on
                continue
            if inst["instruction"] == "swap":
                op1_value = data.loc[feature, inst["op1"]]
                data.loc[feature, inst["op1"]] = data.loc[feature, inst["op2"]]
                data.loc[feature, inst["op2"]] = op1_value
            elif inst["instruction"] == "copy":
                # copy op1 to op2
                data.loc[feature, inst["op2"]] = data.loc[feature, inst["op1"]]
            elif inst["instruction"] == "delete":
                data.loc[feature, inst["op1"]] = np.nan

    return data


def filter_samples_mask(data, smdata, formation_params):
    """
    Returns Boolean Mask of samples to keep
    the data columns must match the sample metadata index
    handles the old way, with 'only_prefs_samples' and 'keep_not_used'
    or the new way, with 'include_types'
    """
    warnings = []
    if not smdata.index.equals(data.columns):
        raise IndexError("Your Sample Metadata does not match your data columns.")

    if "include_types" in formation_params:
        if not formation_params["include_types"]:
            return np.array([True] * len(smdata)), warnings
        to_keep = np.array([False] * len(smdata))
        # Update to_keep for matching injection types
        for inj_type in formation_params["include_types"]:
            to_keep |= smdata["injection_type"].str.startswith(inj_type).values
    else:
        to_keep = np.array([True] * len(smdata))
        # filter out anything not labeled a pref or sample
        if formation_params["only_prefs_samples"]:
            to_keep = (
                smdata["injection_type"].str.startswith("pref").values
                | smdata["injection_type"].str.startswith("sample").values
            )

        # filter out not_used
        if not formation_params["keep_not_used"]:
            not_used = smdata["injection_type"].str.startswith("not_used").values
            to_keep = to_keep & ~not_used
    return to_keep, warnings


def filter_features_mask(data, smdata, fmdata, form_params):
    """
    Generates a filtering boolean key for data
    """
    warnings = []
    to_keep = np.array([True] * len(fmdata))
    if "Primary" in fmdata.columns:
        to_keep = to_keep & fmdata["Primary"].fillna(False)

    if not fmdata.index.equals(data.index):
        raise IndexError("Your Feature Metadata does not match your data index.")

    pref_as = smdata.index[smdata["injection_type"] == "prefa"]
    pref_bs = smdata.index[smdata["injection_type"] == "prefb"]
    miss_column = f"PREF Missing (of {len(pref_as) + len(pref_bs)})"
    if form_params["missing_as_percent"]:
        missing_cutoff = np.ceil(
            (len(pref_as) + len(pref_bs)) * form_params["missing_cutoff"] / 100
        )
    else:
        missing_cutoff = form_params["missing_cutoff"]

    to_keep = to_keep & (fmdata[miss_column] <= missing_cutoff)
    if len(pref_as) > 0:
        to_keep = to_keep & (fmdata["PREFA CVs"] <= form_params["cv_cutoff"] / 100)
    if len(pref_bs) > 0:
        to_keep = to_keep & (fmdata["PREFB CVs"] <= form_params["cv_cutoff"] / 100)

    if "Primary" in fmdata.columns:
        to_keep = to_keep & fmdata["Primary"].fillna(False)

    # annotated compounds don't get filtered by CV or missing PREFs
    to_keep = to_keep | pd.notnull(fmdata["Annotation_ID"])

    # but everything by non_quant
    non_quant = fmdata["Non_Quant"].fillna(False).astype(bool)
    to_keep = to_keep & ~non_quant

    # find all duplicated annotations that weren't filtered as non_quant
    annotated = to_keep & pd.notnull(fmdata["Annotation_ID"])
    is_duplicated = fmdata.loc[annotated, "Annotation_ID"].duplicated(keep=False)
    duplicates = fmdata.loc[is_duplicated.index[is_duplicated]]
    if "Adduct" in fmdata.columns and not form_params["keep_adducts"] == "all":
        if form_params["keep_adducts"] == "top":
            # only keep annotation if top priority adduct
            fmdata_adduct_priority_null_vals = (
                fmdata["Adduct_Priority"].fillna(False).astype(bool)
            )
            fmdata_adduct_priority_defined = fmdata[fmdata_adduct_priority_null_vals]
            fmdata_top_adduct = fmdata_adduct_priority_defined["Adduct_Priority"].map(
                lambda x: x.split(",")[0]
            )
            to_drop = fmdata_adduct_priority_defined["Adduct"] != fmdata_top_adduct
            to_drop_idx = to_drop.index[to_drop]
            to_keep[to_drop_idx] = False
        else:
            for anno in set(duplicates["Annotation_ID"]):
                group = duplicates.loc[duplicates["Annotation_ID"] == anno]
                # pick highest priority adduct, if possible
                priorities = group.loc[group.index[0], "Adduct_Priority"]
                if not priorities:
                    continue
                priorities = priorities.split(",")
                for adduct in priorities:
                    if adduct in group["Adduct"].values:
                        to_drop = group["Adduct"] != adduct
                        to_drop_idx = to_drop.index[to_drop]
                        to_keep[to_drop_idx] = False
                        break

    # find annotations that are still duplicated
    annotated = to_keep & pd.notnull(fmdata["Annotation_ID"])
    is_duplicated = fmdata.loc[annotated, "Annotation_ID"].duplicated(keep=False)
    duplicates = fmdata.loc[is_duplicated.index[is_duplicated]]

    for anno in set(duplicates["Annotation_ID"]):
        group = duplicates.loc[duplicates["Annotation_ID"] == anno]
        # drop QI annotation(s) if there is a TF annotation
        if form_params["feature_priority"][0] in group["__extraction_method"].values:
            to_drop = group["__extraction_method"] == form_params["feature_priority"][1]
            to_drop_idx = to_drop.index[to_drop]
            to_keep[to_drop_idx] = False

    # finally, if there are still duplicates, warn the user
    annotated = to_keep & pd.notnull(fmdata["Annotation_ID"])
    cols_to_check = ["Annotation_ID"]
    if "Adduct" in fmdata.columns and form_params["keep_adducts"] == "all":
        cols_to_check.append("Adduct")
    is_duplicated = fmdata.loc[annotated].duplicated(subset=cols_to_check, keep=False)
    if sum(is_duplicated) > 0:
        warnings.append(
            "There are duplicate annotations that could not be filtered by adduct "
            + "or extraction method. Please remove duplicate annotations before "
            + "sharing results."
        )

    return to_keep.values, warnings


def feature_qc(data, smdata, fmdata):
    """
    Generates QC data and adds to feature metadata
    """
    pref_as = smdata.index[smdata["injection_type"] == "prefa"]
    pref_bs = smdata.index[smdata["injection_type"] == "prefb"]
    miss_column = f"PREF Missing (of {len(pref_as) + len(pref_bs)})"

    if len(pref_as) > 0:
        fmdata["PREFA CVs"] = data.loc[:, pref_as].std(axis=1) / data.loc[
            :, pref_as
        ].mean(axis=1)
    else:
        fmdata["PREFA CVs"] = None
    if len(pref_bs) > 0:
        fmdata["PREFB CVs"] = data.loc[:, pref_bs].std(axis=1) / data.loc[
            :, pref_bs
        ].mean(axis=1)
    else:
        fmdata["PREFB CVs"] = None
    fmdata[miss_column] = data.loc[:, pref_as.union(pref_bs)].isnull().sum(axis=1)
    return fmdata, miss_column


def combine(data, smdata, fmdata, fmdata_formats):
    """Returns a formatted dataset and any warnings"""
    warnings = []
    abundance_threshold = 1
    fmdata = fmdata.copy()
    smdata = smdata.copy()

    # check mdata and data have same index
    dataset_index = data.index
    if set(fmdata.index) != set(data.index):
        warnings.append(
            "Your metadata and data have mismatched features, likely because this"
            " is an old dataset. You might have missing features..."
        )
        dataset_index = fmdata.index.intersection(data.index).copy()
        data = data.loc[dataset_index, :]
        fmdata = fmdata.loc[dataset_index, :]

    # sort and drop index (which was Compound_ID); Compound_ID should be in feature metadata
    fmdata = fmdata.loc[dataset_index, :].reset_index(drop=True)
    data.reset_index(drop=True, inplace=True)
    data[data < abundance_threshold] = np.nan

    # fill non_quants and cast as object
    if "Non_Quant" not in fmdata.columns:
        fmdata["Non_Quant"] = False
    fmdata.fillna({"Non_Quant": False}, inplace=True)
    fmdata["Non_Quant"] = fmdata["Non_Quant"].astype("object")

    # fill values less than 1 with 1 and round
    data = data.clip(lower=1).fillna(0)
    data = data.round(0).astype(np.int64)
    data = data.replace(0, np.nan)

    # keep additional columns but drop columns that are for internal use only
    additional_cols = [
        col for col in fmdata if col not in fmdata_formats and not col.startswith("__")
    ]
    fmdata = fmdata.reindex(columns=additional_cols + list(fmdata_formats))

    final_dataset = pd.concat([fmdata, data], axis="columns")
    final_dataset = _sort_dataset(final_dataset)

    # set headers as top row
    injection_meta = smdata.T
    final_dataset = final_dataset.reset_index(drop=True).reset_index()
    final_dataset = final_dataset.astype("object")
    final_dataset.loc[-1, final_dataset.columns] = final_dataset.columns
    final_dataset.loc[-1, injection_meta.columns] = injection_meta.loc["program_id"]
    final_dataset = final_dataset.sort_index()
    final_dataset = final_dataset.rename(index={-1: "Metabolite"})
    injection_meta = injection_meta.drop(index="program_id")
    injection_meta.loc[:, "Metabolite"] = injection_meta.index.str.capitalize()
    original_columns = final_dataset.columns
    final_dataset = pd.concat([injection_meta, final_dataset], axis="index")
    final_dataset = final_dataset.loc[:, original_columns]
    final_dataset = final_dataset.fillna("")
    return final_dataset, warnings


def to_excel(data, fmdata_formats, sdmata_formats, miss_column, method_name="Default"):
    """
    Returns a BytesIO object of a formatted excel sheet
    """

    output = io.BytesIO()
    workbook = xlsxwriter.Workbook(output)
    worksheet = workbook.add_worksheet(method_name)
    formats = {
        "date": workbook.add_format(
            {
                "font_name": "Arial",
                "font_size": 9,
                "align": "center",
                "num_format": "m/d/yyyy",
            }
        ),
        "int": workbook.add_format(
            {
                "font_name": "Arial",
                "font_size": 9,
                "num_format": "0",
            }
        ),
        "center": workbook.add_format(
            {"font_name": "Arial", "font_size": 9, "align": "center"}
        ),
        "float2": workbook.add_format(
            {"font_name": "Arial", "font_size": 9, "num_format": "0.00"}
        ),
        "float4": workbook.add_format(
            {"font_name": "Arial", "font_size": 9, "num_format": "0.0000"}
        ),
        "label": workbook.add_format(
            {
                "font_name": "Arial",
                "font_size": 9,
                "italic": True,
                "align": "right",
            }
        ),
        "header": workbook.add_format(
            {"font_name": "Arial", "font_size": 9, "bold": True}
        ),
        "header_center": workbook.add_format(
            {"font_name": "Arial", "font_size": 9, "bold": True, "align": "center"}
        ),
        "default": workbook.add_format({"font_name": "Arial", "font_size": 9}),
        "percent": workbook.add_format(
            {"font_name": "Arial", "font_size": 9, "num_format": "0%"}
        ),
    }

    column_formats = {
        header: formats[label] for header, label in fmdata_formats.items()
    }
    row_formats = {index: formats[label] for index, label in sdmata_formats.items()}

    # column widths
    worksheet.set_column_pixels(
        data.columns.get_loc("Compound_ID"),
        data.columns.get_loc("Compound_ID"),
        86,
    )
    worksheet.set_column_pixels(
        data.columns.get_loc("superClass"),
        data.columns.get_loc("superClass"),
        80,
    )
    worksheet.set_column_pixels(
        data.columns.get_loc("mainClass"),
        data.columns.get_loc("mainClass"),
        80,
    )
    worksheet.set_column_pixels(
        data.columns.get_loc("subClass"),
        data.columns.get_loc("subClass"),
        80,
    )
    worksheet.set_column_pixels(
        data.columns.get_loc(miss_column),
        data.columns.get_loc(miss_column),
        80,
    )
    worksheet.set_column_pixels(
        data.columns.get_loc("Annotation_ID"),
        data.columns.get_loc("Annotation_ID"),
        90,
    )
    worksheet.set_column_pixels(
        data.columns.get_loc("Non_Quant"),
        data.columns.get_loc("Non_Quant"),
        80,
    )
    worksheet.set_column_pixels(
        data.columns.get_loc("HMDB_ID"),
        data.columns.get_loc("HMDB_ID"),
        102,
    )
    worksheet.set_column_pixels(
        data.columns.get_loc("HMDB_specificity (1=match; 2=representative)"),
        data.columns.get_loc("HMDB_specificity (1=match; 2=representative)"),
        270,
    )
    worksheet.set_column(
        data.columns.get_loc("Metabolite"),
        data.columns.get_loc("Metabolite"),
        data["Metabolite"].str.len().max(),
    )
    worksheet.set_column_pixels(
        data.columns.get_loc("Metabolite") + 1,
        len(data.columns) - 1,
        150,
    )

    for i, row in enumerate(data.values):
        index_label = data.index[i]
        if index_label in row_formats.keys():  # injection metadata
            worksheet.set_row_pixels(i, 16, row_formats[index_label])
        else:  # default to int
            worksheet.set_row_pixels(i, 16, formats["int"])

        for j, val in enumerate(row):
            column_label = data.columns[j]
            # metadata headers
            if index_label == "Metabolite" and column_label in column_formats.keys():
                worksheet.write(i, j, val, formats["header"])
            # data headers
            elif index_label == "Metabolite":
                worksheet.write(i, j, val, formats["header_center"])
            # row labels
            elif index_label in row_formats.keys() and column_label == "Metabolite":
                worksheet.write(i, j, val, formats["label"])
            # feature metadata
            elif column_label in column_formats.keys():
                worksheet.write(i, j, val, column_formats[column_label])
            else:
                worksheet.write(i, j, val)
    workbook.close()
    return output



# pylint: disable=all
import blueshift

test_data = "../../tests/DCinput1.csv"
test_info = "../../tests/DCinfo1.csv"

############
# Example 1 - Supervised Internal Standard + Nearest Neighbor Drift Correction
############

b = blueshift.DriftCorrection(test_data, test_info)

# output full raw dataset
print(b.data)

# output reference pool file names
print(b.pools)

# perform internal standard correction on rdata using the given internal standard(s)
b.internal_standard_correct(internal_standards="15R-15-methyl PGA2")

# output full drift corrected dataset
print(b.data)

# calculate CVs for IS corrected data
b.calc_cvs()

# output CVs for IS corrected data
print(b.cvs)

# perform nearest neighbor correction on cdata using the given pool
b.pool_correct(interpolation="NN", pool="PREFB", override=False)

# output full drift corrected dataset
print(b.data)

# calculate CVs for all metabolites
b.calculate_cvs()

# save the corrected data and CVs as a csv with the provided filename
b.to_csv("IS-NN_correct.csv")
b.report()

############
# Example 2 - Supervised Linear Drift Correction
############

b = blueshift.DriftCorrection(test_data, test_info)

# output full raw dataset
print(b.data)

# performs the indicated smooth method on cdata (equivalent to rdata until one drift correction method is done)
b.pool_correct(interpolation="linear", override=False, pool="PREFA")

# calculate CVs for all metabolites
b.calculate_cvs()

# save the corrected data and CVs as a csv with a default filename
b.to_csv()
b.report()

############
# Example 3 - Unsupervised Internal Standard + Linear Drift Correction with Override
############

b = blueshift.DriftCorrection(test_data, test_info)

# automatically perform IS correction followed by linear correction with overrides
b.correct(
    internal_standard_kwargs={"internal_standards": "15R-15-methyl PGA2"},
    pool_kwargs={"interpolation": "smooth", "override": True, "pool": "PREFA"},
)

# calculate CVs for all metabolites
b.calculate_cvs()

# output full drift corrected dataset
print(b.data)

# save the corrected data and CVs as a csv with a default filename
b.to_csv()
b.report()



# pylint: disable=redefined-outer-name
"""
Tests for Gravity
"""
from pathlib import Path
import pytest
import pandas as pd
from bmxp import gravity


@pytest.fixture()
def filepath():
    """
    First file path for test
    """
    return Path(__file__).parent / "test_gravity.csv"


@pytest.fixture()
def dataframe(filepath):
    """
    First dataframe for test
    """
    return pd.read_csv(filepath, header=0, index_col="Compound_ID")


def test_initialize_add(dataframe):
    """
    Test MSAligner initializes or throws errors
    """
    gravity.cluster(dataframe)



"""
MSEclipse is a LCMS dataset alignment package, designed to combine datasets which have
been collected on the same method but perhaps on different instruments or columns.
"""

# pylint: disable=too-many-lines
from itertools import chain, combinations
import copy
import logging
import collections
import io
from typing import Optional, List, Union
from itertools import product
import numpy as np
import pandas as pd
import networkx as nx
import textwrap
from scipy import interpolate
import statsmodels.api as sm
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from bmxp import FMDATA

logging.basicConfig()
LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)

np.random.seed(0)

lowess = sm.nonparametric.lowess
__version__ = "0.2.7"


def dataset_loops(attr=None):
    """
    Wraps a function with a double For-loop for iterating over the datasets.
    It converts the supplied datasets, which may be None, 1 dataset, or multiple,
    to a list.
    Also handles creation of a dict key for an attribute and makes it if it doesn't
     exist. E.g. self.scalers[DS1] -- if that key doesn't exist, it will create it
     so that self.scalers[DS1][DS2] can be assigned.
    """

    def attr_handler(func):
        def loops(self, datasets_1=None, datasets_2=None, *args, **kwargs):
            # convert datasets to a list of datasets for the loop
            if datasets_1 is None:
                datasets_1 = list(self.datasets)
            if isinstance(datasets_1, str):
                datasets_1 = [datasets_1]

            if datasets_2 is None:
                datasets_2 = list(self.datasets)
            if isinstance(datasets_2, str):
                datasets_2 = [datasets_2]

            for ds1 in datasets_1:
                if attr == "coarse_matches":
                    LOGGER.info(f"Generating coarse matches for {ds1} to others...")
                if attr == "scalers":
                    LOGGER.info(f"Generating scalers for {ds1} to others...")
                if attr == "scaled_values":
                    LOGGER.info(f"Scaling {ds1} to others...")
                if attr == "stds":
                    LOGGER.info(f"Calculating variance from {ds1} to others...")
                if attr == "matches":
                    LOGGER.info(f"Generating {ds1} feature tables...")
                if attr is None:
                    LOGGER.info(f"Loading feature graph with {ds1} feature tables...")

                if attr and ds1 not in self.__getattribute__(attr):
                    self.__getattribute__(attr)[ds1] = {}
                for ds2 in datasets_2:
                    if ds1 == ds2:
                        continue
                    func(self, ds1, ds2, *args, **kwargs)

        return loops

    return attr_handler


def ds2_loop(func):
    """
    Wraps a function with an inner ds2 loop. Does not check for attribute
     dictionary keys (handled in functions)
    """

    def wrapper(self, ds1=None, ds2=None, *args, **kwargs):
        ds2 = list(self.datasets) if ds2 is None else ds2
        if not isinstance(ds2, str):
            for item in ds2:
                if item == ds1:
                    continue
                func(self, ds1, item, *args, **kwargs)
        else:
            func(self, ds1, ds2, *args, **kwargs)

    return wrapper


class MSAligner:  # pylint: disable=too-many-instance-attributes
    """
    MSAligner stores 2 or more MS Datasets and performs matching, scaling, and
    alignments between them.
    """

    # references FMDATA keys, not schema values
    # values are evaluated in __init__
    descriptors = {"RT": "linear", "MZ": "ppm", "Intensity": "log10"}
    mz_col = "MZ"
    default_cutoff = 6
    default_weight = 1
    default_cutoffs = {}
    default_weights = {}
    default_coarse_params = {
        "RT": {"upper": 0.5, "lower": -0.5},
        "MZ": {"upper": 15, "lower": -15},
        "Intensity": {"upper": 2, "lower": -2},
    }
    default_scaler_params = {
        "smoothing_method": "lowess",
        "smoothing_params": {"frac": 0.1},
    }
    feature_name = "Compound_ID"
    intensity_col = "Intensity"
    annotation_col = "Annotation_ID"

    def __init__(
        self,
        *args: Union[str, pd.DataFrame],
        names: Optional[List[str]] = None,
        schema_labels: Optional[dict[str, str]] = None,
    ):
        fmdata = FMDATA.copy()

        self.mz_col = (schema_labels or {}).get(
            MSAligner.mz_col, FMDATA[MSAligner.mz_col]
        )

        if schema_labels:
            fmdata.update(schema_labels)

        self.descriptors = {fmdata[k]: v for k, v in MSAligner.descriptors.items()}
        self.default_cutoff = MSAligner.default_cutoff
        self.default_weight = MSAligner.default_weight
        self.default_weights = {
            fmdata[k]: v for k, v in MSAligner.default_weights.items()
        }
        self.default_cutoffs = {
            fmdata[k]: v for k, v in MSAligner.default_cutoffs.items()
        }

        # for k in MSAligner.descriptors:
        #     if k not in MSAligner.default_weights:
        if MSAligner.default_weights is None:
            self.default_weights = {
                fmdata[k]: self.default_weight for k in MSAligner.descriptors
            }

        self.default_coarse_params = {
            fmdata[k]: v for k, v in MSAligner.default_coarse_params.items()
        }
        self.default_scaler_params = MSAligner.default_scaler_params
        self.feature_name = fmdata[MSAligner.feature_name]
        self.intensity_col = fmdata[MSAligner.intensity_col]
        self.annotation_col = fmdata[MSAligner.annotation_col]
        self.datasets = collections.OrderedDict()
        self.anchors = {}
        self.coarse_matches = {}
        self.scalers = {}
        self.scaled_values = {}
        self.stds = {}
        self.matches = {}
        self.scores = {}
        self.cutoffs = {}
        self.weights = {}
        self.coarse_params = {}
        self.scaler_params = {}
        self.multipliers = {}
        self.prescalers = {}
        self.graph = nx.DiGraph()
        self.remove_all = True
        # convert names to a list if not supplied
        if names is None:
            names = [None] * len(args)

        # Check there are the same number of datasets as names
        if len(names) != len(args):
            raise ValueError(
                "The number of names you provided does not match the "
                "number of datasets."
            )

        # Add datasets using the add_dataset method
        for arg, name in zip(args, names):
            self.add_dataset(arg, name)

    def add_dataset(self, dataset, name=None):
        """
        Adds a dataset to the existing alignment object.

        :param dataset: str, filepath of .csv dataset
        :param name: str, name of dataset
        """

        if name is None:
            # check if there are any dataframes present; if so, raise an exception
            if isinstance(dataset, pd.DataFrame):
                raise ValueError(
                    "Names must be provided if importing DataFrames directly."
                )
            name = dataset

        if "__" in name:
            raise ValueError(
                "Your dataset name (or filename) cannot contain double underscores."
            )

        if not isinstance(dataset, pd.DataFrame):
            dataset = pd.read_csv(dataset)

        # check that the name isn't already present
        if name in self.datasets.keys():
            raise ValueError("You have duplicate names for your datasets.")

        self.datasets[name] = dataset
        self.datasets[name].index = self.datasets[name].index.map(str)
        self._validate_data(name)

    def _validate_data(self, names=None):
        """
        Performs a check to see if a dataset has any compatibility errors.

        :param names: list, name of all datasets to check
        """

        # If names are not provided, check all
        if names is None:
            names = self.datasets.keys()

        # If one name is provided as a string, convert to a 1 member list
        elif isinstance(names, str):
            names = [names]

        # Check all datasets included in the names
        for name in names:
            ds = self.datasets[name]
            # calculate intensity if needed
            if (
                self.intensity_col not in ds.columns
                and self.intensity_col in self.descriptors
            ):
                gen_intensity(
                    ds,
                    self.intensity_col,
                    ignore=[
                        descr
                        for descr in self.descriptors
                        if descr != self.intensity_col
                    ]
                    + [self.feature_name],
                )

            # Check required columns
            required_columns = [*self.descriptors, self.feature_name]
            for req_col in required_columns:
                if req_col not in ds.columns:
                    raise ValueError(f"Column {req_col} not found in dataset {name}.")

    def set_instance_defaults(self, params):
        """
        Updates default values for weights and cutoffs.
        """

        for key, item in params.items():
            if key.lower() == "weight":
                self.default_weight = item
            elif key.lower() == "cutoff":
                self.default_cutoff = item
            elif key.lower() == "weights":
                self.default_weights.update(item)
            elif key.lower() == "cutoffs":
                self.default_cutoffs.update(item)
            elif key.lower() == "coarse_params":
                self.default_coarse_params.update(item)
            elif key.lower() == "scaler_params":
                self.default_scaler_params.update(item)
            else:
                raise KeyError(
                    f"Unknown value: {key}. Acceptable values are "
                    " 'weight', 'weights', 'cutoff', 'cutoffs',"
                    " 'coarse_params' and 'scaler_params'."
                )

    def set_params(self, params, rec=True):
        """
        Sets weights, cutoffs, and scalers for any dataset pairs
        For example:
        {"scalers": {
            "DS1": {"DS2": {"RT": 1}}
            },
        "cutoffs": { ...}
        }
        """
        for key, item in params.items():
            if key.lower() == "scalers":
                self.set_scalers(item, rec)
            elif key.lower() == "cutoffs":
                self.set_cutoffs(item, rec)
            elif key.lower() == "weights":
                self.set_weights(item, rec)
            elif key.lower() == "coarse_params":
                self.set_coarse_params(item, rec)
            elif key.lower() == "scaler_params":
                self.set_scaler_params(item, rec)
            else:
                raise KeyError(
                    f"Unknown value: {key}. Acceptable values are 'scalers', "
                    "'cutoffs', 'weights', 'coarse_params', and 'scaler_params'."
                )

    def set_scalers(self, scalers, rec=True):
        """
        Function to set custom scalers.
         If rec is true, will calculate the reciprocal scalers
         scalers is a nested dict with the dataframe names
         and scalers to be used. For example:
        {
            'DS1': {
                'DS2':{
                    'RT': {'x': [1, 2, 3], 'y': [.1, .2, .3]},
                    'MZ': {'x': [100, 200, 300], 'y' : [2, 4, 3]},
                    'Intensity': 1.001
                }
            }
        }
        While it seems clunky, it allows us to set multiple scalers at once.
         If rec is True, and recipricol scalers are specified, they will be overwritten.
        :param scalers: dict
        :param rec: bool
        """
        for ds1 in scalers:
            # make scaler dicts if they don't exist
            if ds1 not in self.scalers:
                self.scalers[ds1] = {}
            for ds2 in scalers[ds1]:
                if ds2 not in self.scalers[ds1]:
                    self.scalers[ds1][ds2] = {}
                if rec:
                    if ds2 not in self.scalers:
                        self.scalers[ds2] = {}
                    if ds1 not in self.scalers[ds2]:
                        self.scalers[ds2][ds1] = {}

                # Check if dataframes are in the aligner object
                if ds1 not in self.datasets or ds2 not in self.datasets:
                    raise IndexError(
                        f"Either {ds1} or {ds2} was not found in the datasets."
                    )

                # Check for typos in scaler names
                if not set(scalers[ds1][ds2].keys()).issubset(
                    set(self.descriptors.keys())
                ):
                    raise IndexError(
                        "The only acceptable columns for scalers are "
                        f"{self.descriptors.keys()}. "
                        f"You provided {list(scalers.keys())}."
                    )

                self._set_scalers(scalers[ds1][ds2], ds1, ds2, rec)

    def _set_scalers(self, scalers, ds1, ds2, rec):
        """
        Takes a list of scalers in a dictionary form
         and sets them as scalers in the object.
        For example:
        {
            'RT': {'x': [1, 2, 3], 'y': [.1, .2, .3]},
            'MZ': {'x': [100, 200, 300], 'y' : [2, 4, 3]},
            'Intensity': {'x': [3, 4, 5], 'y' : [.1, .3, .4]}
        }
        Scalers should already be in their proper space

        :param scalers: Dict(str: Dict(str:list),
         containing the scaler names as Keys and values
        :param ds1: str, name of Source dataset
        :param ds2: str, name of Target dataset
        :param rec: bool, Flag to create reciprocal or not (DS1->DS2, DS2->DS1)
            Convenience for manually specified scalers, not for use autogenerated
        """
        for k, v in scalers.items():
            values = list(zip(*v))
            df = pd.DataFrame({"x": values[0], "y": values[1]})
            self.scalers[ds1][ds2][k] = df
            if rec:
                if self.descriptors[k] in ["linear", "log10"]:
                    self.scalers[ds2][ds1][k] = pd.DataFrame(
                        {"x": df["x"] + df["y"], "y": -df["y"]}
                    )
                elif self.descriptors[k] == "ppm":
                    new_xs = df["x"] + df["x"] * df["y"] / 1000000
                    new_ys = -df["y"] * df["x"] / new_xs
                    self.scalers[ds2][ds1][k] = pd.DataFrame({"x": new_xs, "y": new_ys})
                else:
                    raise NotImplementedError(
                        "Not sure how to reverse scale the mode ", self.descriptors[k]
                    )

    def set_prescalers(self, prescalers):
        """
        Dict containing datasets and descriptors providing the dataset and
        the instructions to be applied.
        Five spread across the gradient should be plenty.
        This will shift DS1 RTs earlier, since the instructions are negative
        {
            'DS1': {
                'RT': [[1, 5, 10], [0, -0.2, -0.3]]
            }
        }
        """
        self.prescalers = {}
        for dataset in prescalers:
            self.prescalers[dataset] = {}
            for descriptor in prescalers[dataset]:
                if descriptor not in self.descriptors:
                    raise KeyError(
                        f"Your prescaler descriptor {descriptor} is not present as a "
                        "descriptor."
                    )
                self.prescalers[dataset][descriptor] = [
                    pd.Series(prescalers[dataset][descriptor][0]),
                    pd.Series(prescalers[dataset][descriptor][1]),
                ]

    def set_cutoffs(self, cutoffs, rec=True):
        """
        Function to set custom cutoffs. If rec is true, apply reciprocal cutoffs.
        {
            'DS1': {
                'DS2':{
                    'RT': 5.5,
                    'MZ': 5.5,
                    'Intensity': 5.0
                }
            }
        }
        :param cutoffs: dict
        :param rec: bool
        """
        cutoffs = cutoffs.copy()
        for ds1 in cutoffs:
            # make cutoff dicts if they don't exist
            if ds1 not in self.cutoffs:
                self.cutoffs[ds1] = {}
            for ds2 in cutoffs[ds1]:
                if ds2 not in self.cutoffs[ds1]:
                    self.cutoffs[ds1][ds2] = {}

                # Check if dataframes are in the aligner object
                if ds1 not in self.datasets or ds2 not in self.datasets:
                    raise IndexError(
                        f"Either {ds1} or {ds2} was not found in the datasets."
                    )

                # Check for typos in cutoff names
                if not set(cutoffs[ds1][ds2].keys()).issubset(set(self.descriptors)):
                    raise IndexError(
                        "The only acceptable columns for cutoffs are "
                        f"{self.descriptors.keys()}. You provided "
                        f"{cutoffs[ds1][ds2].keys()}."
                    )

                # replace 0's with inf, negating their use in cutoffs
                for key, value in cutoffs[ds1][ds2].items():
                    if value == 0:
                        cutoffs[ds1][ds2][key] = float("inf")

                self.cutoffs[ds1][ds2].update(cutoffs[ds1][ds2])
                if rec:
                    if ds2 not in self.cutoffs:
                        self.cutoffs[ds2] = {}
                    if ds1 not in self.cutoffs[ds2]:
                        self.cutoffs[ds2][ds1] = {}
                    self.cutoffs[ds2][ds1].update(cutoffs[ds1][ds2])

    def set_weights(self, weights, rec=True):
        """
        Function to set custom weights. If rec is true, apply reciprocal weights
        {
            'DS1': {
                'DS2':{
                    'RT': .5,
                    'MZ': 1,
                    'Intensity': 0
                }
            }
        }
        :param weights: dict
        :param rec: bool
        """
        weights = weights.copy()
        for ds1 in weights:
            # make weight dicts if they don't exist
            if ds1 not in self.weights:
                self.weights[ds1] = {}
            for ds2 in weights[ds1]:
                if ds2 not in self.weights[ds1]:
                    self.weights[ds1][ds2] = {}

                # Check if dataframes are in the aligner object
                if ds1 not in self.datasets or ds2 not in self.datasets:
                    raise IndexError(
                        f"Either {ds1} or {ds2} was not found in the datasets."
                    )

                # Check for typos in weight names
                if not set(weights[ds1][ds2].keys()).issubset(set(self.descriptors)):
                    raise IndexError(
                        "The only acceptable columns for weights are "
                        f"{self.descriptors.keys()}.  You provided "
                        f"{weights[ds1][ds2].keys()}."
                    )

                self.weights[ds1][ds2].update(weights[ds1][ds2])
                if rec:
                    if ds2 not in self.weights:
                        self.weights[ds2] = {}
                    if ds1 not in self.weights[ds2]:
                        self.weights[ds2][ds1] = {}
                    self.weights[ds2][ds1].update(weights[ds1][ds2])

    def set_coarse_params(
        self, coarse_params, rec=True
    ):  # pylint: disable=too-many-branches
        """
        Function to set custom cutoffs. If rec is true, apply reciprocal cutoffs
        {
            'DS1': {
                'DS2':{
                    'rt_minus': -.7,
                    'ppm_plus': 4
                }
            }
        }
        :param coarse_params: dict
        :param rec: bool
        """
        coarse_params = coarse_params.copy()
        for ds1 in coarse_params:
            # make coarse_params dicts if they don't exist
            if ds1 not in self.coarse_params:
                self.coarse_params[ds1] = {}
            for ds2 in coarse_params[ds1]:
                if ds2 not in self.coarse_params[ds1]:
                    self.coarse_params[ds1][ds2] = {}

                # Check if dataframes are in the aligner object
                if ds1 not in self.datasets or ds2 not in self.datasets:
                    raise IndexError(
                        f"Either {ds1} or {ds2} was not found in the datasets."
                    )

                # Check for typos in coarse_param names
                if not set(coarse_params[ds1][ds2].keys()).issubset(
                    set(key for key in self.descriptors)
                ):
                    raise IndexError(
                        f"The only acceptable columns for coarse_params are"
                        f" {set(key for key in self.descriptors)}. You"
                        f" provided {coarse_params[ds1][ds2].keys()}."
                    )

                self.coarse_params[ds1][ds2].update(coarse_params[ds1][ds2])
                if rec:
                    rec_dict = {}
                    for descriptor in coarse_params[ds1][ds2]:
                        new_feature = {}
                        if "upper" in coarse_params[ds1][ds2][descriptor]:
                            new_feature["lower"] = -coarse_params[ds1][ds2][descriptor][
                                "upper"
                            ]
                        if "lower" in coarse_params[ds1][ds2][descriptor]:
                            new_feature["upper"] = -coarse_params[ds1][ds2][descriptor][
                                "lower"
                            ]
                        rec_dict.update({descriptor: new_feature})

                    if ds2 not in self.coarse_params:
                        self.coarse_params[ds2] = {}
                    if ds1 not in self.coarse_params[ds2]:
                        self.coarse_params[ds2][ds1] = {}
                    self.coarse_params[ds2][ds1].update(rec_dict)

    def set_scaler_params(self, scaler_params, rec=True):
        """
        Function to set custom smoothing params for scaler generation. If rec is true,
        apply reciprocal params
        {
            'DS1': {
                'DS2':{
                    'smoothing_method': 'lowess',
                    'smoothing_params': {
                        'frac': 0.9
                    }
                }
            }
        }
        :param scaler_params: dict
        :param rec: bool
        """
        scaler_params = scaler_params.copy()
        for ds1 in scaler_params:
            # make scaler_params dicts if they don't exist
            if ds1 not in self.scaler_params:
                self.scaler_params[ds1] = {}
            for ds2 in scaler_params[ds1]:
                if ds2 not in self.scaler_params[ds1]:
                    self.scaler_params[ds1][ds2] = {}

                # Check if dataframes are in the aligner object
                if ds1 not in self.datasets or ds2 not in self.datasets:
                    raise IndexError(
                        f"Either {ds1} or {ds2} was not found in the datasets."
                    )

                self.scaler_params[ds1][ds2].update(scaler_params[ds1][ds2])
                if rec:
                    if ds2 not in self.scaler_params:
                        self.scaler_params[ds2] = {}
                    if ds1 not in self.scaler_params[ds2]:
                        self.scaler_params[ds2][ds1] = {}
                    self.scaler_params[ds2][ds1].update(scaler_params[ds1][ds2])

    def to_csv(
        self,
        *args,
        filepath=None,
        to_bytes=False,
        c_size_or_loss=0,
        g_size_or_loss=0,
        diameter=1,
        remove_rerank=True,
    ):
        """
        Condenses network into a flat CSV file
        :param filepath, str, location to be saved
        :param *args, str, names of datasets that you want in the export file
        :param, g_size_or_loss, int, minimum size of groups
            (positive, or #datasets - if 0 or negative)
        :param, c_size_or_loss, int, minimum size of clique
            (positive, or #datasets - if 0 or negative)
        :param, diameter, int, specifies maximum distance to be in a group
        :param, rank_and_remove, bool, True - removes nodes after adding to a group,
            False - produces all combinations that aren't subsets of larger groups.
        """

        if len(args) < 1:
            targ_datasets = list(self.datasets)
        else:
            targ_datasets = list(args)
        df = self.results(
            *targ_datasets,
            c_size_or_loss=c_size_or_loss,
            g_size_or_loss=g_size_or_loss,
            diameter=diameter,
            remove_rerank=remove_rerank,
        )
        if to_bytes:
            bytes_csv = df.to_csv().encode()
            return io.BytesIO(bytes_csv)
        if filepath is None:
            filepath = "results.csv"
        df.to_csv(filepath)
        return None

    def results(
        self,
        *args,
        c_size_or_loss=0,
        g_size_or_loss=0,
        diameter=1,
        remove_rerank=True,
    ):
        """
        Returns a results dataframe
        """
        if len(args) < 1:
            targ_datasets = list(self.datasets)
        else:
            targ_datasets = list(args)

        df = self.deconvolute(
            targ_datasets,
            remove_rerank,
            c_size_or_loss,
            g_size_or_loss,
            diameter,
        )
        for ds in df.columns:
            df = df.merge(
                self.datasets[ds],
                how="left",
                left_on=ds,
                right_index=True,
                suffixes=(None, f"_{ds}"),
            )
        return df

    @staticmethod
    def compress_graph(graph):
        """
        Compresses network into single highest rank non-directed graph
        """
        graph = graph.copy()
        selected_edges = [(u, v) for u, v, e in graph.edges(data=True) if e["rank"] > 0]
        for edge in selected_edges:
            graph.remove_edge(*edge)
        compressed = graph.to_undirected(reciprocal=True)

        # clear edge weights and update
        for u, v, d in compressed.edges(data=True):
            compressed[u][v]["score"] = 0
        for u, v, d in graph.edges(data=True):
            if u in compressed.neighbors(v):
                compressed[u][v]["score"] += d["score"]
        return compressed

    @staticmethod
    def get_rank_groups(graph, max_size, c_size=None, g_size=None, diameter=1):
        """
        Given a subgraph, ranks all possible groups based on provided info.
        Returns a list of ({group members}, clique_size, group_size,
        number of edges, avg edge weight)
        Ranking is accomplished by sorting by group size, clique size,
        number of edges, then edge weight (lowest to highest)
        """

        if c_size is None:
            c_size = max_size

        if g_size is None:
            g_size = max_size

        # some sanity checks
        if c_size > g_size:
            raise UserWarning(
                "You have indicated a clique size minimum greater than the group size m"
                "inimum. This is likely unintentional."
            )

        if max_size + 1 < c_size + diameter:
            raise UserWarning(
                "Your diameter and clique size setting is likely a mistake. If you use "
                "a diameter of 2, set 'c_size_or_loss' to -1, or any number less than t"
                "he number of datasets. For a diameter of 3, set 'c_size_or_loss' to -2"
                " or any number that is 2 below the number of datasets."
            )

        # use cliques as seeds
        cliques = set(
            frozenset(clique)
            for clique in nx.find_cliques(graph)
            if len(clique) >= c_size
        )

        potential_groups = set()

        if diameter == 1:
            potential_groups = cliques
        else:
            for clique in cliques:
                if diameter == 2:
                    # Try adding each nodes neighbors
                    neighbors = {
                        node: set(graph.neighbors(node)) - clique for node in clique
                    }
                    potential_groups.update(
                        set(
                            frozenset(clique.union(neighbors[node]))
                            for node in clique
                            if len(clique.union(neighbors[node])) >= g_size
                        )
                    )
                if diameter == 3:
                    # Initialize valid_nodes as a defaultdict of sets
                    valid_nodes = collections.defaultdict(set)

                    # Populate valid_nodes with neighbors grouped by partite
                    for clique_node in clique:
                        for neighbor_node in set(graph.neighbors(clique_node)) - clique:
                            partite = neighbor_node.split("__")[0]
                            valid_nodes[partite].add(neighbor_node)
                    target_g_size = sum(
                        1 for partite in valid_nodes.values() if partite
                    ) + len(clique)
                    if target_g_size < g_size:
                        continue
                    # add all frozensets to potential groups
                    non_empty = [nodes for nodes in valid_nodes.values() if nodes]
                    potential_groups.update(
                        set(
                            frozenset(combination).union(frozenset(clique))
                            for combination in product(*non_empty)
                        )
                    )

        # remove subsets
        filtered_groups = []
        for g1 in potential_groups:
            for g2 in potential_groups:
                if g1 == g2:
                    continue
                if g1.issubset(g2):
                    break
            else:
                filtered_groups.append(g1)

        # score and return
        final_groups = []
        for g in filtered_groups:
            sg = graph.subgraph(g)
            # score and record
            if (sg.number_of_edges()) == 0:
                weight = 0
            else:
                weight = (
                    sum(data["score"] for _, _, data in sg.edges(data=True))
                    / sg.number_of_edges()
                )
            max_clique_size = 0
            for clique in nx.find_cliques(sg):
                if len(clique) > max_clique_size:
                    max_clique_size = len(clique)
            final_groups.append(
                (set(g), len(sg), max_clique_size, len(sg.edges()), weight)
            )
        final_groups = sorted(final_groups, key=lambda x: (-x[1], -x[2], -x[3], x[4]))
        return final_groups

    def deconvolute(
        self,
        datasets,
        remove_rerank=True,
        c_size_or_loss=0,
        g_size_or_loss=0,
        diameter=1,
    ):
        """
        Makes a dataframe where each row is a feature, and each column is a dataset
        Entry is null if there is no matching clique
        """
        if g_size_or_loss <= 0:
            g_size_or_loss = len(self.datasets) + g_size_or_loss
        if c_size_or_loss <= 0:
            c_size_or_loss = len(self.datasets) + c_size_or_loss

        # compressed = self.compress_graph(self.graph)
        df_indices = {ds: [] for ds in self.datasets}
        results = []
        for dsg in nx.weakly_connected_components(self.graph):
            # skip graphs smaller than size
            if len(dsg) < g_size_or_loss:
                continue
            compressed = self.compress_graph(self.graph.subgraph(dsg))
            for sg in nx.connected_components(compressed):
                sg = compressed.subgraph(sg).copy()
                sg_results = []
                if remove_rerank:
                    while True:
                        new = self.get_rank_groups(
                            sg,
                            len(self.datasets),
                            c_size_or_loss,
                            g_size_or_loss,
                            diameter,
                        )
                        if len(new) == 0:
                            break
                        sg_results.append(new[0])
                        sg.remove_nodes_from(new[0][0])
                else:
                    sg_results = self.get_rank_groups(
                        sg,
                        len(self.datasets),
                        c_size_or_loss,
                        g_size_or_loss,
                        diameter,
                    )

                # remove those which don't have a dataset we're interested in
                ds_filtered = []
                for g in sg_results:
                    group_ds = [feature_name.split("__")[0] for feature_name in g[0]]
                    if len(set(datasets).intersection(group_ds)) >= 1:
                        ds_filtered.append(g)

                results.extend(ds_filtered)

        # parse results
        for row in results:
            # add an empty row fore everything, then fill in with features
            for df in df_indices.keys():
                df_indices[df].append(None)
            for feature in row[0]:
                ds_name, feature_name = feature.split("__")
                df_indices[ds_name][-1] = feature_name
        return pd.DataFrame.from_dict(df_indices)

    def align(self, ds1=None, ds2=None):
        """
        Performs all steps for generating the network
        """
        np.seterr(divide="ignore")
        self.gen_anchors(ds1, ds2)
        self.gen_coarse(ds1, ds2)
        self.gen_scalers(ds1, ds2)
        self.gen_scaled_values(ds1, ds2)
        self.gen_stds(ds1, ds2)
        self.gen_matches(ds1, ds2)
        self.gen_graph(ds1, ds2)
        np.seterr(divide="warn")

    def prescale(self):
        """
        Prescales all datasets
        """
        for ds in self.datasets:
            if ds not in self.prescalers:
                continue
            df = self.datasets[ds]
            for descriptor in self.descriptors:
                if descriptor not in self.prescalers[ds]:
                    continue
                prescalers = self.prescalers[ds][descriptor]
                try:
                    descriptor_col = df.columns.get_loc(descriptor)
                except KeyError as e:
                    raise KeyError(
                        f"The descriptor {descriptor} was not found in dataset {ds}."
                    ) from e
                if f"{descriptor}_Original" not in df.columns:
                    df.insert(
                        descriptor_col + 1, f"{descriptor}_Original", df[descriptor]
                    )
                df[descriptor] = scale_to(
                    df[f"{descriptor}_Original"],
                    prescalers[0],
                    prescalers[1],
                    self.descriptors[descriptor],
                )

    def gen_anchors(self, ds1=None, ds2=None):
        """
        Creates and sets anchors for all datasets specified.
        If either ds1 or ds2 is None, calculates anchors fall all datasets.
        """
        LOGGER.info("Generating anchors...")
        self.prescale()
        if ds1 is None or ds2 is None:
            datasets = self.datasets
        else:
            datasets = set()
            for ds in [ds1, ds2]:
                if isinstance(ds, str):
                    datasets.add(ds)
                else:
                    datasets.update(ds)
        coarse_params = copy.deepcopy(
            {k: self.default_coarse_params[k] for k in self.descriptors}
        )

        # copy the mode into the params
        for descr_param in coarse_params:
            coarse_params[descr_param].update({"mode": self.descriptors[descr_param]})

        for ds in datasets:
            if self.mz_col in set(self.descriptors.keys()):
                self.anchors[ds] = anchors_mz(
                    self.datasets[ds],
                    coarse_params,
                    self.mz_col,
                    self.remove_all,
                )
            else:
                self.anchors[ds] = anchors(
                    self.datasets[ds], coarse_params, self.remove_all
                )

    @dataset_loops("coarse_matches")
    def gen_coarse(self, ds1, ds2):
        """
        Generates all coarse anchors between datasets
        """
        coarse_params = self.default_coarse_params.copy()
        try:
            coarse_params.update(self.coarse_params[ds1][ds2])
        except KeyError:
            pass

        # remove coarse params which aren't in the descriptors
        coarse_params = copy.deepcopy({k: coarse_params[k] for k in self.descriptors})
        for descr_param in coarse_params:
            coarse_params[descr_param].update({"mode": self.descriptors[descr_param]})
        # extract anchors, match, and convert results to a dataframe
        if self.mz_col in self.descriptors.keys():
            res = simple_match_mz(
                self.datasets[ds1].loc[self.anchors[ds1], :],
                self.datasets[ds2].loc[self.anchors[ds2], :],
                coarse_params,
                self.mz_col,
            )
        else:
            res = simple_match(
                self.datasets[ds1].loc[self.anchors[ds1], :],
                self.datasets[ds2].loc[self.anchors[ds2], :],
                coarse_params,
            )
        self.coarse_matches[ds1][ds2] = pd.DataFrame(res)
        self.coarse_matches[ds1][ds2].columns = [ds1, ds2]

    @dataset_loops("scalers")
    def gen_scalers(self, ds1, ds2):
        """
        Creates and saves scalers for rt, mz, and intensity.
        These autogenerated scalers will not overwrite scalers already in place.
        """
        scaler_params = self.default_scaler_params.copy()
        try:
            scaler_params.update(self.scaler_params[ds1][ds2])
        except KeyError:
            pass
        coarse_match = self.coarse_matches[ds1][ds2]
        ds1_matches = self.datasets[ds1].loc[coarse_match[ds1], :]
        ds2_matches = self.datasets[ds2].loc[coarse_match[ds2], :]
        if ds2 not in self.scalers[ds1]:
            self.scalers[ds1][ds2] = {}

        scalers = {}
        for descr, mode in self.descriptors.items():
            if descr not in self.scalers[ds1][ds2]:
                scaled = calc_scalers(
                    ds1_matches[descr].values,
                    ds2_matches[descr].values,
                    smoothing=scaler_params["smoothing_method"],
                    mode=mode,
                    **scaler_params["smoothing_params"],
                )
                scalers[descr] = zip(*scaled)
        self._set_scalers(scalers, ds1, ds2, rec=False)

    @dataset_loops("scaled_values")
    def gen_scaled_values(self, ds1, ds2):
        """
        Scales the rt, mz, and intensity values for all datasets
        """
        # extract anchors, match, and convert results to a dataframe
        scaled = pd.DataFrame(index=self.datasets[ds1].index)
        scalers = self.scalers[ds1][ds2]
        for descr, mode in self.descriptors.items():
            scaled[descr] = scale_to(
                self.datasets[ds1][descr],
                scalers[descr]["x"],
                scalers[descr]["y"],
                mode=mode,
            )
        self.scaled_values[ds1][ds2] = scaled

    @dataset_loops("stds")
    def gen_stds(self, ds1, ds2):
        """
        Calculates standard deviations to be used for matching.
        """
        coarse_matches = self.coarse_matches[ds1][ds2]
        scalers = self.scalers[ds1][ds2]
        self.stds[ds1][ds2] = {}
        for descr, mode in self.descriptors.items():
            ds1_vals = self.datasets[ds1].loc[coarse_matches[ds1], descr]
            ds2_vals = self.datasets[ds2].loc[coarse_matches[ds2], descr]
            s_ds1_vals = scale_to(
                ds1_vals, scalers[descr]["x"], scalers[descr]["y"], mode=mode
            )
            if mode == "linear":
                res = (s_ds1_vals - ds2_vals).values
            if mode == "ppm":
                res = ((s_ds1_vals - ds2_vals) / s_ds1_vals * 1_000_000).values
            if mode == "log10":
                res = (np.log10(s_ds1_vals) - np.log10(ds2_vals)).values

            self.stds[ds1][ds2][descr] = np.std(res[~is_outlier(res)])

    @dataset_loops("matches")
    def gen_matches(self, ds1, ds2):
        """
        Generates best matches for each dataset
        """
        # generate the weights and cutoffs
        # if no "custom" cutoffs or weights are present, use default
        cutoffs = {}
        weights = {}
        for descr in self.descriptors:
            try:
                cutoffs[descr] = self.cutoffs[ds1][ds2][descr]
            except KeyError:
                try:
                    cutoffs[descr] = self.default_cutoffs[descr]
                except KeyError:
                    cutoffs[descr] = self.default_cutoff
            try:
                weights[descr] = self.weights[ds1][ds2][descr]

            except KeyError:
                try:
                    weights[descr] = self.default_weights[descr]
                except KeyError:
                    weights[descr] = self.default_weight
        if ds1 not in self.scores:
            self.scores[ds1] = {}

        if self.mz_col in self.descriptors.keys():
            self.matches[ds1][ds2], self.scores[ds1][ds2] = score_match_mz(
                self.scaled_values[ds1][ds2],
                self.datasets[ds2],
                self.stds[ds1][ds2],
                self.descriptors,
                cutoffs,
                weights,
                self.mz_col,
            )
        else:
            self.matches[ds1][ds2], self.scores[ds1][ds2] = score_match(
                self.scaled_values[ds1][ds2],
                self.datasets[ds2],
                self.stds[ds1][ds2],
                self.descriptors,
                cutoffs,
                weights,
            )

    @dataset_loops()
    def gen_graph(self, ds1, ds2):
        """
        Generates the network graph to be used for feature selection
        """
        nodes = [f"{ds1}__{str(i)}" for i in self.datasets[ds1].index]
        self.graph.add_nodes_from(nodes)
        matches = self.matches[ds1][ds2]
        scores = self.scores[ds1][ds2]
        edges = []
        for col in matches:
            this_match = matches[col][~pd.isnull(matches[col])]
            edge_scores = scores[col][~pd.isnull(matches[col])].values
            sources = this_match.index
            sources = [ds1 + "__" + str(name) for name in sources]
            targets = this_match.values
            targets = [ds2 + "__" + str(name) for name in targets]
            edges.extend(
                [
                    tuple([s, t, {"rank": col, "score": score}])
                    for s, t, score in zip(sources, targets, edge_scores)
                ]
            )
        self.graph.add_edges_from(edges)

    def report(
        self,
        datasets_1=None,
        datasets_2=None,
        filepath="report.pdf",
        to_bytes=False,
    ):
        """
        Creates a report of the alignment parameters.
        """
        if datasets_1 is None:
            datasets_1 = list(self.datasets)
        if isinstance(datasets_1, str):
            datasets_1 = [datasets_1]

        if datasets_2 is None:
            datasets_2 = list(self.datasets)
        if isinstance(datasets_2, str):
            datasets_2 = [datasets_2]
        if to_bytes:
            file_handle = io.BytesIO()
        else:
            file_handle = filepath

        # format plot text
        (
            title_wrap_width,
            x_label_wrap_width,
            y_label_wrap_width,
            plot_padding,
        ) = (
            55,
            65,
            40,
            1,
        )

        # create scaling report
        with PdfPages(file_handle) as pdf:
            for ds1 in datasets_1:
                for ds2 in datasets_2:
                    if (
                        ds1 == ds2
                        or ds1 not in self.scalers
                        or ds2 not in self.scalers[ds1]
                    ):
                        continue
                    for results in [False, True]:
                        if results and (
                            ds1 not in self.matches or ds2 not in self.matches[ds1]
                        ):
                            continue
                        plt.figure(figsize=(24, 14))
                        eval_chart(self, ds1, ds2, show=False, results=results)
                        for ax in plt.gcf().get_axes():
                            ax.set_title(
                                "\n".join(
                                    textwrap.wrap(ax.get_title(), title_wrap_width)
                                ),
                                fontsize=10,
                            )
                            ax.set_xlabel(
                                "\n".join(
                                    textwrap.wrap(ax.get_xlabel(), x_label_wrap_width)
                                ),
                                fontsize=8,
                                labelpad=5,
                            )
                            ax.set_ylabel(
                                "\n".join(
                                    textwrap.wrap(ax.get_ylabel(), y_label_wrap_width)
                                ),
                                fontsize=8,
                                labelpad=5,
                            )
                        plt.tight_layout(pad=plot_padding)
                        pdf.savefig()
                        plt.close("all")
        if to_bytes:
            file_handle.seek(0)
            return file_handle
        return None

    def explain(
        self,
        feature_tuples=None,
        annotation_ids=None,
        show_plot=True,
        label_edges=False,
        compress=True,
    ):
        """
        Creates a directed subgraph of feature's weakly interconnected nodes

        """
        combined_subgraph = nx.DiGraph()
        target_nodes = []
        if feature_tuples:
            if len(feature_tuples) == 2 and isinstance(feature_tuples[0], str):
                feature_tuples = [feature_tuples]
            for dataset, compound_id in feature_tuples:
                df = self.datasets[dataset]
                cmp_idx = df.loc[
                    df[self.feature_name] == compound_id, self.feature_name
                ].index[0]
                node_name = f"{dataset}__{cmp_idx}"

                # Find the target node
                target_node = None
                for node, data in self.graph.nodes(data=True):
                    if node == node_name:
                        target_node = node
                        target_nodes.append(target_node)
                        break
                if target_node is None:
                    continue

                component_subgraph = self.get_component(self.graph, target_node)
                if compress:
                    component_subgraph = self.compress_graph(component_subgraph)
                component_subgraph = component_subgraph.to_directed()
                combined_subgraph = nx.compose(combined_subgraph, component_subgraph)

        if annotation_ids:
            if isinstance(annotation_ids, str):
                annotation_ids = [annotation_ids]
            for dataset_name, df in self.datasets.items():
                matching_indices = df[
                    df[self.annotation_col].isin(annotation_ids)
                ].index
                for idx in matching_indices:
                    node_name = f"{dataset_name}__{idx}"
                    if node_name in self.graph:
                        target_nodes.append(node_name)
                        component_subgraph = self.get_component(self.graph, node_name)
                        if compress:
                            component_subgraph = self.compress_graph(component_subgraph)

                        # Convert to directed graph if not already
                        component_subgraph = component_subgraph.to_directed()
                        combined_subgraph = nx.compose(
                            combined_subgraph, component_subgraph
                        )
        node_labels = {}
        for node in combined_subgraph:
            [n_dataset, n_idx] = node.split("__")
            try:
                n_annotation = self.datasets[n_dataset].loc[n_idx, self.annotation_col]
            except:
                n_annotation = ""

            if pd.isnull(n_annotation):
                n_annotation = ""
            n_cmp_id = self.datasets[n_dataset].loc[n_idx, self.feature_name]

            node_labels[node] = f"{n_dataset}-{n_cmp_id}\n{n_annotation}"

        self.plot_subgraph(
            combined_subgraph,
            show_plot=show_plot,
            label_edges=label_edges,
            node_labels=node_labels,
            selected_nodes=target_nodes,
        )
        return combined_subgraph

    def get_component(self, graph, target_node):
        """
        Returns a subgraph of a target node's weakly interconnected nodes
        """
        visited_nodes = set()
        # Traverse outgoing edges
        for u, v in nx.bfs_edges(graph, source=target_node):
            visited_nodes.add(u)
            visited_nodes.add(v)
        # Traverse incoming edges
        for u, v in nx.bfs_edges(graph.reverse(), source=target_node):
            visited_nodes.add(u)
            visited_nodes.add(v)
        component_subgraph = graph.subgraph(visited_nodes).copy()
        return component_subgraph

    def plot_subgraph(
        self,
        subgraph,
        filepath="subgraph_plot.pdf",
        to_bytes=False,
        label_edges=True,
        show_plot=True,
        node_labels=None,
        selected_nodes=None,
    ):
        """
        Plots the subgraph.
        """
        if to_bytes:
            file_handle = io.BytesIO()
        else:
            file_handle = filepath

        plt.figure(figsize=(8, 6))
        pos = nx.spring_layout(subgraph)

        shape_cycle = [
            "o",  # Circle
            "^",  # Triangle
            "s",  # Square
            "p",  # Pentagon
            "*",  # Star
            "h",  # Hexagon
            "D",  # Diamond
            "d",  # Thin diamond
            "P",  # Plus
            "X",  # Cross
            "8",  # Figure 8
        ]
        color_cycle = [
            "red",
            "green",
            "blue",
            "orange",
            "purple",
        ]
        dataset_markers = {}

        for node in subgraph.nodes():
            dataset, _ = node.split("__")
            if dataset not in dataset_markers:
                dataset_markers[dataset] = [
                    shape_cycle[len(dataset_markers) % len(shape_cycle)],
                    color_cycle[len(dataset_markers) % len(color_cycle)],
                ]

        selected_node_list = []
        non_selected_node_list = []
        node_styles = {}

        for node in subgraph.nodes():
            dataset, _ = node.split("__")
            shape, color = dataset_markers[dataset]
            node_styles[node] = {"shape": shape, "color": color}

            if selected_nodes and node in selected_nodes:
                selected_node_list.append(node)
            else:
                non_selected_node_list.append(node)

        # Draw non-selected nodes
        for dataset, (shape, color) in dataset_markers.items():
            nodes_with_style = [
                node
                for node in non_selected_node_list
                if node.split("__")[0] == dataset
            ]
            if nodes_with_style:
                nx.draw_networkx_nodes(
                    subgraph,
                    pos,
                    nodelist=nodes_with_style,
                    node_color="white",
                    edgecolors=color,
                    node_shape=shape,
                    node_size=300,
                    linewidths=1.5,
                )

        # Draw selected nodes
        for dataset, (shape, color) in dataset_markers.items():
            nodes_with_style = [
                node for node in selected_node_list if node.split("__")[0] == dataset
            ]
            if nodes_with_style:
                nx.draw_networkx_nodes(
                    subgraph,
                    pos,
                    nodelist=nodes_with_style,
                    node_color=color,
                    edgecolors=color,
                    node_shape=shape,
                    node_size=300,
                    linewidths=1.5,
                )
        # Draw edges
        bidirectional_edges = []
        monodirectional_edges = []

        for u, v in subgraph.edges():
            if subgraph.has_edge(v, u):
                bidirectional_edges.append((u, v))
            else:
                monodirectional_edges.append((u, v))
        nx.draw_networkx_edges(
            subgraph,
            pos,
            edgelist=bidirectional_edges,
            edge_color="black",
            arrowstyle="->",
        )
        nx.draw_networkx_edges(
            subgraph,
            pos,
            edgelist=monodirectional_edges,
            edge_color="lightgray",
            style="dashed",
            arrowstyle="->",
        )
        # Draw labels
        label_pos = {node: (x + 0.02, y + 0.02) for node, (x, y) in pos.items()}
        nx.draw_networkx_labels(subgraph, label_pos, labels=node_labels, font_size=8)

        # Draw edge labels
        if label_edges:
            edge_labels = nx.get_edge_attributes(subgraph, "score")
            edge_labels = {w: f"{edge_labels[w]:.2f}" for w in edge_labels}
            nx.draw_networkx_edge_labels(
                subgraph, pos, edge_labels=edge_labels, font_size=6
            )
        # Create legend
        legend_elements = [
            plt.Line2D(
                [0],
                [0],
                marker=style[0],
                color=style[1],
                label=dataset,
                markerfacecolor="white",
                markeredgecolor=style[1],
                markeredgewidth=1.5,
                markersize=8,
            )
            for dataset, style in dataset_markers.items()
        ] + [
            plt.Line2D(
                [0],
                [0],
                marker="o",
                color="black",
                label="Selected Nodes",
                markerfacecolor="black",
                markeredgewidth=1.5,
                markersize=8,
            )
        ]
        plt.legend(
            handles=legend_elements,
            title="Datasets",
            loc="best",
            bbox_to_anchor=(1.0, 0.5),
            fontsize="small",
            title_fontsize="small",
        )
        plt.title("Directed Subgraph")
        plt.tight_layout()
        plt.savefig(file_handle, format="pdf", bbox_inches="tight")
        if show_plot:
            plt.show()
        plt.close()
        if to_bytes:
            file_handle.seek(0)
            return file_handle
        return None


def calc_scalers(x1, x2, smoothing=None, mode="linear", **kwargs):
    """
    Creates and returns scalers from X1 and X2 coordinates
    i.e. Dataset 1 RTs - [7.1, 8.43, 9.5]
         Dataset 2 RTs - [7.24, 8.74, 9.7]
    Returns scalers in the proper space, e.g. log10 space for intensity, ppm for mz
         Dataset 1 RTs -     [7.1, 8.43, 9.5]
         Dataset 2 Scalers - [0.14, 0.31, 0.2]

    :param x1: Array or List
    :param x2: Array or List
    :param smoothing: str, Smoothing mode (None or "lowess" currently)
    :param mode: str, "linear" or "ppm"
    :return: Numpy array
    """
    x1 = np.array(x1)
    x2 = np.array(x2)

    if mode == "linear":
        y = x2 - x1
    elif mode == "ppm":
        y = (x2 - x1) / x1 * 1000000
    elif mode == "log10":
        y = np.log10(x2) - np.log10(x1)
        x1 = np.log10(x1)
    else:
        raise NotImplementedError(f"We do not have a scaling method for {mode}.")
    if smoothing is None:
        return (x1, y)

    # +/- <0.5 ppt random noise to break ties
    np.random.seed(0)
    noise = (np.random.random(size=len(x1)) - 0.5) * x1 / 10**12

    # allow us to override frac via kwargs
    temp_kwargs = kwargs.copy()
    if "frac" not in temp_kwargs:
        temp_kwargs["frac"] = 0.1

    delta = 0.01 * (max(x1) - min(x1))
    scalers = lowess(y, x1 + noise, return_sorted=False, delta=delta, **temp_kwargs)

    # On failure, depending on statsmodels version, either returns NaNs, or
    # the residuals. Try adding more until it works or reaches 1. On reaching 1,
    # it should be obvious that smoothing failed or it is over smoothed.
    while np.isnan(scalers).all() or np.array_equal(scalers, y):
        temp_kwargs["frac"] += 0.01
        if temp_kwargs["frac"] >= 1.0:
            break
        scalers = lowess(
            y,
            x1 + noise,
            return_sorted=False,
            delta=delta,
            **temp_kwargs,
        )
    return (x1, scalers)


def score_match(df1, df2, stds, descriptors, cutoffs, weights, top_n=1):
    """
    Returns scored best matches for a dataframe to another
    :param df1: DataFrame, first dataset
    :param df2: DataFrame, second dataset
    :param stds: Dict, standard deviations for "RT", "MT", and "Intensity"
    :param cutoffs: Dict, Multipliers for finding cutoffs
    :param weights: Dict, Multipliers for weights
    :param top_n: Int, number of potential matches to record
    :return: Dataframe, df1 index as index, and matching df2 index in top_n columns
    """
    results_indices = []
    results_scores = []

    np1 = df1[list(descriptors)].values
    np2 = df2[list(descriptors)].values

    for row in np1:
        potential_hits = np.full(len(np2), True, dtype=bool)
        ds1_values = {}

        # generate values and bounds
        for i, (descr, mode) in enumerate(descriptors.items()):
            ds1_values[descr] = row[i]
            if mode == "linear":
                bounds = (
                    ds1_values[descr] - cutoffs[descr] * stds[descr],
                    ds1_values[descr] + cutoffs[descr] * stds[descr],
                )
            elif mode == "ppm":
                bounds = (
                    ds1_values[descr]
                    - (ds1_values[descr] / 1000000 * cutoffs[descr]) * stds[descr],
                    ds1_values[descr]
                    + (ds1_values[descr] / 1000000 * cutoffs[descr]) * stds[descr],
                )
            elif mode == "log10":
                bounds = (
                    10 ** (np.log10(ds1_values[descr]) - cutoffs[descr] * stds[descr]),
                    10 ** (np.log10(ds1_values[descr]) + cutoffs[descr] * stds[descr]),
                )
            else:
                raise NotImplementedError("Not sure how to handle the mode ", mode)

            # whittle hits down to list
            potential_hits = potential_hits & (
                (np2[:, i] >= bounds[0]) & (np2[:, i] <= bounds[1])
            )

        hits_index = df2.index[potential_hits]

        # score and sort results
        scores = []
        for hit in np2[potential_hits]:
            ds2_values = {descr: hit[i] for i, descr in enumerate(list(descriptors))}
            scores.append(
                calc_score(ds1_values, ds2_values, descriptors, stds, weights)
            )

        # append the top n
        this_hit = []
        this_score = []
        for score, hit in sorted(zip(scores, hits_index))[:top_n]:
            this_score.append(score)
            this_hit.append(hit)
        results_scores.append(this_score)
        results_indices.append(this_hit)

    return pd.DataFrame(results_indices, index=df1.index), pd.DataFrame(
        results_scores, index=df1.index
    )


def anchors_mz(ds, match_params, mz_key, remove_all=True, priority="Intensity"):
    """
    Shortcut for anchors when a descriminating column like MZ is present.
    Allows for binary search
    :param ds: DataFrame
    :param match_params: dict, parameters for matching
    :param mz_key: str, descriptor for binary search
    :param remove_all: bool, flag to indicate removing all confusable feature
    :param priority: str, the column indicating the which features are highest quality
        i.e. higher intensity is a strong feature
    :return: list, the index names of the anchors
    """
    thresholds = {k: {} for k in match_params if k != mz_key}
    if not remove_all:
        columns = list(set(match_params.keys()) | {priority})
        ds = ds[columns].copy()
    else:
        ds = ds[[*match_params.keys()]].copy()
    ds["index"] = ds.index
    ds = ds.sort_values(by=mz_key).reset_index(drop=True)

    for descriptor in thresholds:
        thresholds[descriptor]["values"] = ds[descriptor].to_numpy()
        tol = (
            abs(match_params[descriptor]["upper"] - match_params[descriptor]["lower"])
            / 2
        )

        if match_params[descriptor]["mode"] == "log10":
            thresholds[descriptor]["lower"] = thresholds[descriptor]["values"] / 10**tol
            thresholds[descriptor]["upper"] = thresholds[descriptor]["values"] * 10**tol
        if match_params[descriptor]["mode"] == "linear":
            thresholds[descriptor]["lower"] = thresholds[descriptor]["values"] - tol
            thresholds[descriptor]["upper"] = thresholds[descriptor]["values"] + tol
        if match_params[descriptor]["mode"] == "ppm":
            thresholds[descriptor]["lower"] = thresholds[descriptor]["values"] - (
                thresholds[descriptor]["values"] / 1e6 * tol
            )
            thresholds[descriptor]["upper"] = thresholds[descriptor]["values"] + (
                thresholds[descriptor]["values"] / 1e6 * tol
            )

    mz_array = ds[mz_key].to_numpy()
    orig_index = ds["index"].to_numpy()
    ppm_tol = abs(match_params[mz_key]["upper"] - match_params[mz_key]["lower"]) / 2
    mz_upper = mz_array + mz_array * ppm_tol / 1e6
    mz_lower = mz_array - mz_array * ppm_tol / 1e6
    graph = nx.Graph()
    graph.add_nodes_from(orig_index)

    for i in range(len(mz_array)):
        j_start = np.searchsorted(mz_array, mz_lower[i], side="left")
        j_end = np.searchsorted(mz_array, mz_upper[i], side="right")
        if j_start >= j_end:
            continue
        hits = np.array([True] * (j_end - j_start))
        for descriptor in thresholds:
            candidates = thresholds[descriptor]["values"][j_start:j_end]
            hits = (
                hits
                & (candidates > thresholds[descriptor]["lower"][i])
                & (candidates < thresholds[descriptor]["upper"][i])
            )
        for j in np.flatnonzero(hits) + j_start:
            if j == i:
                continue
            graph.add_edge(orig_index[i], orig_index[j])
    to_keep = []
    for component in nx.connected_components(graph):
        if len(component) == 1:
            to_keep.append(next(iter(component)))
        elif not remove_all:
            component_list = list(component)
            max_idx = ds.loc[component_list, "Intensity"].idxmax()
            to_keep.append(max_idx)
    to_keep.sort()
    return to_keep


def score_match_mz(df1, df2, stds, descriptors, cutoffs, weights, mz_key, top_n=1):
    """
    Shortcut for score_match when a discriminating column, like "MZ" is present.
    Allows for a fast binary search
    Returns scored best matches for a dataframe to another
    :param df1: DataFrame, first dataset
    :param df2: DataFrame, second dataset
    :param stds: Dict, standard deviations for "RT", "MT", and "Intensity"
    :param cutoffs: Dict, Multipliers for finding cutoffs
    :param weights: Dict, Multipliers for weights
    :param mz_key: Str, descriptor label for binary search
    :param top_n: Int, number of potential matches to record
    :return: Dataframe, df1 index as index, and matching df2 index in top_n columns
    """
    results_indices = []
    results_scores = []

    other_keys = [k for k in descriptors if k != mz_key]
    df1 = df1[[mz_key] + other_keys].copy()
    df2 = df2[[mz_key] + other_keys].copy()
    df2 = df2.sort_values(by=mz_key)

    mz1 = df1[mz_key].to_numpy()
    mz2 = df2[mz_key].to_numpy()
    index2 = df2.index.to_numpy()

    mz_tol = cutoffs[mz_key] * stds[mz_key]
    mz_lower = mz1 - mz1 * mz_tol / 1e6
    mz_upper = mz1 + mz1 * mz_tol / 1e6

    # Prepare descriptor values and bounds
    vals1 = {}
    vals2 = {}
    for k in other_keys:
        v1 = df1[k].to_numpy()
        v2 = df2[k].to_numpy()
        tol = cutoffs[k] * stds[k]
        mode = descriptors[k]
        if mode == "log10":
            lower = v1 / 10**tol
            upper = v1 * 10**tol
        elif mode == "linear":
            lower = v1 - tol
            upper = v1 + tol
        elif mode == "ppm":
            lower = v1 - (v1 * tol / 1e6)
            upper = v1 + (v1 * tol / 1e6)
        else:
            raise ValueError(f"Unknown mode for {k}")
        vals1[k] = {"v": v1, "lo": lower, "hi": upper}
        vals2[k] = v2

    for i in range(len(mz1)):
        mz_i = mz1[i]
        j_start = np.searchsorted(mz2, mz_lower[i], side="left")
        j_end = np.searchsorted(mz2, mz_upper[i], side="right")

        if j_start >= j_end:
            results_indices.append([None] * top_n)
            results_scores.append([None] * top_n)
            continue

        # Apply descriptor filters
        hits = np.ones(j_end - j_start, dtype=bool)
        for k in other_keys:
            candidates = vals2[k][j_start:j_end]
            hits &= (candidates >= vals1[k]["lo"][i]) & (
                candidates <= vals1[k]["hi"][i]
            )

        if not np.any(hits):
            results_indices.append([None] * top_n)
            results_scores.append([None] * top_n)
            continue

        candidates_mz = mz2[j_start:j_end][hits]
        idx = index2[j_start:j_end][hits]
        score = ((((mz_i - candidates_mz) / mz_i) * 1e6) / stds[mz_key]) ** 2 * weights[
            mz_key
        ]
        for k in other_keys:
            v1 = vals1[k]["v"][i]
            v2 = vals2[k][j_start:j_end][hits]
            if descriptors[k] == "linear":
                score += ((v1 - v2) / stds[k]) ** 2 * weights[k]
            elif descriptors[k] == "log10":
                score += (np.log10(v1 / v2) / stds[k]) ** 2 * weights[k]
            elif descriptors[k] == "ppm":
                score += ((v1 - v2) / v1 * 1e6 / stds[k]) ** 2 * weights[k]

        top_idx = np.argsort(score)[:top_n]
        score_top = score[top_idx]
        index_top = idx[top_idx]

        pad_len = top_n - len(score_top)
        results_indices.append(index_top.tolist() + [None] * pad_len)
        results_scores.append(score_top.tolist() + [None] * pad_len)

    return (
        pd.DataFrame(results_indices, index=df1.index),
        pd.DataFrame(results_scores, index=df1.index),
    )


def simple_match_mz(df1, df2, match_params, mz_key):
    """
    Binary search match based on a discriminating mz_key with
    additional descriptor filtering.

    :param df1: DataFrame, source
    :param df2: DataFrame, target
    :param match_params: dict, format:
        {
            "MZ": {"lower": -5, "upper": 5, "mode": "ppm"},
            "RT": {"lower": -0.2, "upper": 0.2, "mode": "linear"},
            ...
        }
    :param mz_key: str, primary key used for binary search
    :return: list of [df1_index, ds2_index]
    """
    other_keys = [k for k in match_params if k != mz_key]
    df1 = df1[[mz_key] + other_keys].copy()
    df2 = df2[[mz_key] + other_keys].copy()
    df2 = df2.sort_values(by=mz_key)

    mz1 = df1[mz_key].to_numpy()
    mz2 = df2[mz_key].to_numpy()

    mz_lower = mz1 + mz1 * match_params[mz_key]["lower"] / 1e6
    mz_upper = mz1 + mz1 * match_params[mz_key]["upper"] / 1e6

    vals1 = {}
    vals2 = {}
    for k in other_keys:
        v1 = df1[k].to_numpy()
        v2 = df2[k].to_numpy()
        mode = match_params[k]["mode"]
        if mode == "log10":
            lower = v1 * 10 ** match_params[k]["lower"]
            upper = v1 * 10 ** match_params[k]["upper"]
        elif mode == "linear":
            lower = v1 + match_params[k]["lower"]
            upper = v1 + match_params[k]["upper"]
        elif mode == "ppm":
            lower = v1 + (v1 * match_params[k]["lower"] / 1e6)
            upper = v1 + (v1 * match_params[k]["upper"] / 1e6)
        else:
            raise ValueError(f"Unknown mode for {k}")
        vals1[k] = {"v": v1, "lo": lower, "hi": upper}
        vals2[k] = v2

    results = []
    for i in range(len(mz1)):
        j_start = np.searchsorted(mz2, mz_lower[i], side="left")
        j_end = np.searchsorted(mz2, mz_upper[i], side="right")
        if j_start >= j_end:
            continue

        hits = np.ones(j_end - j_start, dtype=bool)
        for k in other_keys:
            candidates = vals2[k][j_start:j_end]
            hits &= (candidates > vals1[k]["lo"][i]) & (candidates < vals1[k]["hi"][i])

        for m in df2.index[j_start:j_end][hits]:
            results.append([df1.index[i], m])
    return results


def anchors(ds, match_params, remove_all=True, priority="Intensity"):
    """
    Creates a list of anchors for a given dataset.
    :param ds: DataFrame
    :param match_params: dict, parameters for matching
    :param remove_all: bool, flag to indicate removing all confusable feature
    :param priority: str, the column indicating the which features are highest quality
        i.e. higher intensity is a strong feature
    :return: list, the index names of the anchors
    """
    if not remove_all:
        columns = list(set(match_params.keys()) | {priority})
        ds = ds[columns]
        ds = ds.sort_values(priority, ascending=False)
    else:
        ds = ds[[*match_params.keys()]]

    # convert to numpy for performance
    index = ds.index
    ds = ds.values

    inclusion = set()
    exclusion = set()
    for i, row in enumerate(ds):
        # if a feature has been marked for removal, skip it.
        # important for non-remove all algorithm
        if not remove_all and index[i] in exclusion:
            continue
        bool_array = np.full(len(index), True, dtype=bool)
        for j, param in enumerate(match_params.values()):
            descr_val = row[j]
            tol = abs(param["upper"] - param["lower"]) / 2
            if param["mode"] == "linear":
                lower = descr_val - tol
                upper = descr_val + tol
            elif param["mode"] == "ppm":
                lower = descr_val - (descr_val / 1_000_000 * tol)
                upper = descr_val + (descr_val / 1_000_000 * tol)
            elif param["mode"] == "log10":
                descr_val = float(descr_val)
                lower = descr_val / 10**tol
                upper = descr_val * 10**tol
            bool_array = bool_array & (ds[:, j] > lower) & (ds[:, j] < upper)

        results = ds[bool_array]
        result_indices = index[bool_array]
        if remove_all:
            if len(results) == 1:
                inclusion.add(index[i])
            elif len(results) > 1:
                # if there is more than one result, exclude all of them
                exclusion.update(set(result_indices))
        else:
            if len(results) == 1:
                # if the only result is itself add it and remove the others
                inclusion.add(index[i])
            elif len(results) > 1:
                if result_indices[0] == index[i]:
                    inclusion.add(index[i])
                    not_max = np.delete(result_indices, 0, 0)
                    exclusion.update(set(not_max))

    ds_anchors = list(inclusion - exclusion)
    ds_anchors.sort()
    return ds_anchors


def simple_match(df1, df2, match_params):
    """
    Simply iterates through rows and finds features that fall within a
     threshold in a reference dataset.
    In the future this function will be much more generalized
    :param df1: DataFrame, first dataset
    :param df2: DataFrame, second dataset
    :param match_params: dict, match parameters
    :return: list, 2D Nested containing indices for df1 and df2 matches
    """
    results = []

    np1 = df1[[*match_params.keys()]].values
    np2 = df2[[*match_params.keys()]].values
    for i, row in enumerate(np1):
        bool_array = np.full(len(df2), True, dtype=bool)
        for j, param in enumerate(match_params.values()):
            descr_val = row[j]
            if param["mode"] == "linear":
                lower = descr_val + param["lower"]
                upper = descr_val + param["upper"]
            elif param["mode"] == "ppm":
                lower = descr_val + (descr_val / 1_000_000 * param["lower"])
                upper = descr_val + (descr_val / 1_000_000 * param["upper"])
            elif param["mode"] == "log10":
                lower = descr_val * 10 ** param["lower"]
                upper = descr_val * 10 ** param["upper"]
            bool_array = bool_array & (np2[:, j] > lower) & (np2[:, j] < upper)

        for result in df2.index[bool_array]:
            results.append([df1.index[i], result])

    return results


def gen_intensity(dataset, intensity_col="Intensity", ignore=None):
    """
    calculate an intensity column, using all columns that are not named '
     RT','MZ','Compound_ID'
    overwrites the existing dataframe with another one where Intensity is at index 1

    :param dataset: pandas Dataframe where there are intensities to calculate
    """
    if ignore:
        calc_df = dataset.drop(columns=ignore)
    calc_df.fillna(0, inplace=True)
    dataset.insert(1, intensity_col, calc_df.mean(axis=1, numeric_only=True))


def scale_to(x, x_scalers, y_scalers, mode="linear"):
    """
    Adjusts the x values to the scalers provided.

    :param x: Numpy Array, unscaled values
    :param x_scalers: Numpy Array, x values for the scalers
    :param y_scalers: Numpy Array, y values for the scalers
    :return: Numpy Array, scaled values for x
    """
    if mode == "linear":
        pass
    elif mode == "ppm":
        y_scalers = y_scalers * x_scalers / 1_000_000
    elif mode == "log10":
        x = np.log10(x)
    else:
        raise NotImplementedError(f"We do not have a scaling method for {mode}.")

    x_min_y = y_scalers.iloc[np.argmin(x_scalers)]
    x_max_y = y_scalers.iloc[np.argmax(x_scalers)]
    scaled = x + interpolate.interp1d(
        x_scalers, y_scalers, bounds_error=False, fill_value=(x_min_y, x_max_y)
    )(x)
    if isinstance(scaled, pd.Series):
        scaled = scaled.values
    if mode == "log10":
        scaled = 10**scaled
    return scaled


def is_outlier(points, thresh=3.5):
    """
    Returns a boolean array with True if points are outliers and False
    otherwise.
    Borrowed from https://stackoverflow.com/questions/22354094/pythonic-way-of-detecting
    -outliers-in-one-dimensional-observation-data

    Boris Iglewicz and David Hoaglin (1993), "Volume 16: How to Detect and
        Handle Outliers", The ASQC Basic References in Quality Control:
        Statistical Techniques, Edward F. Mykytka, Ph.D., Editor.

    :param points: An numobservations by numdimensions array of observations
    :param thresh: The modified z-score to use as a threshold. Observations with
    :return: Numpy Array, A numobservations-length boolean array.
    """
    try:
        points = points.values
    except AttributeError:
        pass
    median = np.median(points, axis=0)
    diff = (points - median) ** 2
    diff = np.sqrt(diff)
    med_abs_deviation = np.median(diff)
    modified_z_score = 0.6745 * diff / med_abs_deviation
    return modified_z_score > thresh


def calc_score(v1, v2, descriptors, stds, weights):
    """
    Returns the score of two features, provided their parameters, standard deviations,
     and scoring weights
    :return: float, score
    """
    penalty = 0
    for descr, mode in descriptors.items():
        if mode == "linear":
            penalty += ((v1[descr] - v2[descr]) / stds[descr]) ** 2 * weights[descr]
        elif mode == "ppm":
            penalty += (
                (((v1[descr] - v2[descr]) / v1[descr]) * 1_000_000) / stds[descr]
            ) ** 2 * weights[descr]
        elif mode == "log10":
            penalty += (np.log10(v1[descr] / v2[descr]) / stds[descr]) ** 2 * weights[
                descr
            ]
        else:
            raise NotImplementedError("Not sure how to score mode ", mode)
    return penalty


def eval_chart(self, key1, key2, show=True, results=False):
    """
    Description
    """

    ds1 = self.datasets[key1]
    ds2 = self.datasets[key2]

    if not results:
        coarse_df = self.coarse_matches[key1][key2]
        ds1 = ds1.loc[coarse_df[key1]]
        ds2 = ds2.loc[coarse_df[key2]]
        title = "Coarse"
    else:
        # filter our datasets to just the matched features
        has_match = pd.notnull(self.matches[key1][key2][0])
        matches = self.matches[key1][key2].loc[has_match]
        ds1_matches = matches.index
        ds2_matches = matches[0]
        ds1 = ds1.loc[ds1_matches]
        ds2 = ds2.loc[ds2_matches]
        title = "Matched"

    plt.figure(figsize=(20, 12))
    for i, (descr, mode) in enumerate(self.descriptors.items()):
        descr_1 = ds1[descr].values
        descr_2 = ds2[descr].values
        scalers = self.scalers[key1][key2][descr].sort_values(by="x")

        # calculate coarse scalers, or pull matched scaled values
        if not results:
            s_descr_1 = scale_to(descr_1, scalers["x"], scalers["y"], mode=mode)
        else:
            s_descr_1 = self.scaled_values[key1][key2].loc[ds1_matches][descr]
        if mode == "linear":
            y_vals = descr_2 - descr_1
            s_y_vals = descr_2 - s_descr_1
        if mode == "ppm":
            y_vals = 1e6 * (descr_2 - descr_1) / descr_1
            s_y_vals = 1e6 * (descr_2 - s_descr_1) / s_descr_1
        if mode == "log10":
            y_vals = np.log10(descr_2) - np.log10(descr_1)
            s_y_vals = np.log10(descr_2) - np.log10(s_descr_1)
            descr_1 = np.log10(descr_1)

        max_length = 50
        if len(key1) > max_length:
            label1 = f"...{key1[-(max_length - 3):]}"
        else:
            label1 = key1
        if len(key2) > max_length:
            label2 = f"...{key2[-(max_length - 3):]}"
        else:
            label2 = key2

        # plot the deltas
        plt.subplot(len(self.descriptors), 3, 3 * i + 1)
        plt.scatter(
            descr_1,
            y_vals,
            alpha=0.4,
            s=3,
            rasterized=True,
        )
        plt.title(f"{descr} {label1} vs {label2}")
        plt.xlabel(f"{title} {descr} {label1}")
        plt.ylabel(f"{title} {descr} ({label2} - {label1}) : {mode}")
        plt.plot(scalers["x"], scalers["y"], color="red", rasterized=True)

        # plot the scaled
        plt.subplot(len(self.descriptors), 3, 3 * i + 2)
        plt.scatter(
            descr_1,
            s_y_vals,
            alpha=0.4,
            s=3,
            rasterized=True,
        )
        plt.title(f"{descr} {label1} vs {label2}")
        plt.xlabel(f"{title} {descr} {label1}")
        plt.ylabel(f"{title} {descr} ({label2} - {label1}) : {mode}")
        plt.axhline(y=0, color="red", rasterized=True)

        # plot histogram
        plt.subplot(len(self.descriptors), 3, 3 * i + 3)
        plt.hist(
            x=(s_y_vals),
            bins="auto",
            alpha=0.7,
            rwidth=0.85,
        )
        plt.title(f"{descr} Histograms")
        plt.xlabel(f"{title} {descr} ({label2} - {label1})")
        plt.tight_layout()

    if show:
        plt.show()



"""
A quick example on how to use Gravity
"""

import pandas as pd
from bmxp.gravity import cluster

# load a dataset
df = pd.read_csv("../../tests/test_gravity.csv", header=0, index_col="Compound_ID")

# cluster, with an RT cutoff of 0.025 and required correlation of .9
sample_df = cluster(df, 0.025, 0.9)

# returns a dataframe with Cluster_Num and Cluster_Size
print(df)

# add to our own dataframe
df["Cluster_Num"] = sample_df["Cluster_Num"]
df["Cluster_Size"] = sample_df["Cluster_Size"]

# record results
df.to_csv("results.csv")



"""
Correlation calculations in python; a fallback in case of a missing binary file or
unsupported operating system
"""

import numpy as np
from scipy import stats


def backfill(a, b):
    """Backfill NaNs in each array based on the non-NaN values in the other. wraps
    _backfill for more convenient use

    :param a: array-like, first data array
    :param b: array-like, second data array
    """
    a, b = np.nan_to_num((a, b))
    a_filled = _backfill(a, b)
    b_filled = _backfill(b, a)  # pylint: disable=arguments-out-of-order

    a_filled[a_filled == 0] = np.nan
    b_filled[b_filled == 0] = np.nan

    return a_filled, b_filled


def _backfill(a, b):
    """Backfill NaNs in a based on non-NaN values in b

    :param a: array-like, first data array, array to backfill
    :param b: array-like, second data array
    """
    zeroes = len(a) - np.count_nonzero(a)
    fill_limit = min(np.count_nonzero(a), np.count_nonzero(b))
    if zeroes == 0 or fill_limit == 0:  # nothing to fill
        return a

    sorted_index = np.lexsort((b, a))
    fill_value = -1

    for i in range(zeroes, max(zeroes - fill_limit, 0), -1):
        index = sorted_index[i - 1]
        if b[index] == 0:  # out of non-NaN values
            break
        a[index] = fill_value
        fill_value -= 1

    return a


def free_p(_):
    """Empty function to match function name of C implementation"""


def pearson(x, y, _):
    """Pearson correlation with NaNs dropped

    :param x: array-like, first data array
    :param y: array-like, second data array
    """
    not_nan = np.logical_and(~np.isnan(x), ~np.isnan(y))
    x = x[not_nan]
    y = y[not_nan]
    return np.corrcoef(x, y)[0][1]


def spearman(x, y, _, drop_nan, legacy_mode):
    """Spearman correlation calculation

    :param x: array-like, first data array
    :param y: array-like, second data array
    :param drop_nan: boolean, when False, missing values are backfilled
    :param legacy_mode: boolean, when True, NaNs are filled with 0s when correlating
    features with very few overlapping non-NaN values
    """
    if legacy_mode and np.logical_and(~np.isnan(x), ~np.isnan(y)).sum() < 10:
        # ensure we don't modify the original array
        x = np.copy(x)
        y = np.copy(y)
        x[np.isnan(x)] = 0
        y[np.isnan(y)] = 0
    if not drop_nan:
        x, y = backfill(x, y)
    not_nan = np.logical_and(~np.isnan(x), ~np.isnan(y))
    x = x[not_nan]
    y = y[not_nan]

    # this is about 2x faster than scipy.stats.spearmanr
    rank_x = stats.rankdata(x)
    rank_y = stats.rankdata(y)
    return np.corrcoef(rank_x, rank_y)[0][1]


def pearson_array(arr, r1_start, r2_start, size):
    """Pearson method that accepts a 2D array and relevant indexes

    :param arr: 2D array-like, full set of data
    :param r1_start: int, index of first data array
    :param r2_start: int, index of second data array
    :param size: unused
    """
    return pearson(arr[r1_start], arr[r2_start], size)


def spearman_array(arr, r1_start, r2_start, size, drop_nan, legacy_mode):
    """Spearman method that accepts a 2D array and relevant indexes

    :param arr: 2D array-like, full set of data
    :param r1_start: int, index of first data array
    :param r2_start: int, index of second data array
    :param size: unused
    :param drop_nan: boolean, when False, missing values are backfilled
    :param legacy_mode: boolean, when True, NaNs are filled with 0s when correlating
    features with very few overlapping non-NaN values
    """
    return spearman(arr[r1_start], arr[r2_start], size, drop_nan, legacy_mode)



"""
blueshift is a LCMS dataset drift correction package, designed to correct for drift or
sudden shifts in signal intensity, as measured in pooled reference samples or internal
standards.
"""

from warnings import warn
import re
import pandas as pd
import numpy as np
from tqdm import tqdm
from scipy import interpolate
from bmxp import FMDATA, IMDATA, POOL_INJ_TYPES

__version__ = "0.2.4"


class DriftCorrection:  # pylint: disable=too-many-instance-attributes
    """
    DriftCorrection stores a dataset and corresponding sample information sheet and
    corrects sample signals using pooled reference sample or internal standard-based
    correction methods.
    """

    def __init__(
        self,
        raw_data,
        sample_information,
        default_met_col=None,
        schema_labels=None,
        pool_inj_types=None,
    ):
        """
        Initializes the DriftCorrection object with a dataset and corresponding sample
        information sheet.
        :param raw_data: DataFrame or str, dataset or filepath to .csv dataset
        :param sample_information: DataFrame or str, sample information or filepath to
        .csv sample information
        :param pool_names: tuple[str], pooled reference sample names
        """
        fmdata = FMDATA.copy()
        imdata = IMDATA.copy()
        if schema_labels is not None:
            fmdata.update(schema_labels)
            imdata.update(schema_labels)
        if pool_inj_types is None:
            pool_inj_types = POOL_INJ_TYPES

        self.data = pd.DataFrame()
        self.rdata = pd.DataFrame()
        self.metadata = pd.DataFrame()
        self.sample_info = pd.DataFrame()
        self.pools = {}
        self.batches = {}
        self.correction_params = {}

        # column labels
        self.injection_id = imdata["injection_id"]
        self.injection_type = imdata["injection_type"]
        self.column_number = imdata["column_number"]
        self.injection_order = imdata["injection_order"]
        self.batches_label = imdata["batches"]
        self.QCRole = imdata["QCRole"]
        self.Non_Quant = fmdata["Non_Quant"]
        self.Batches_Skipped = fmdata["Batches Skipped"]

        if default_met_col:
            self.met_col = default_met_col
        else:
            self.met_col = fmdata["Metabolite"]

        if not isinstance(raw_data, pd.DataFrame):
            raw_data = pd.read_csv(raw_data)

        if not isinstance(sample_information, pd.DataFrame):
            sample_information = pd.read_csv(sample_information)

        raw_data, sample_information, self.first_sample = self._validate_data(
            raw_data, sample_information
        )

        self.data, self.metadata = self._parse_data(raw_data)
        self.rdata = self.data.copy()
        self.cvs = pd.DataFrame(index=self.rdata.index)
        self.sample_info = sample_information
        self.sample_info[self.batches_label] = find_batch_start_end(
            self.sample_info,
            injection_order_label=self.injection_order,
            column_number_label=self.column_number,
            batches_label=self.batches_label,
            qc_role_label=self.QCRole,
        )
        self.pools, self.not_used_pools = self._find_pools(pool_inj_types)
        self.batches["default"], self.batches["override"] = self._gen_batches()

    def _validate_data(self, raw_data, sample_information):
        """
        Checks that the dataset and sample information have valid headers and contents
        :param raw_data: DataFrame, raw dataset
        :param sample_information: DataFrame or str, sample information or filepath to
        .csv sample information
        """
        # required columns in sample information
        required_columns = [
            self.injection_id,
            self.injection_type,
            self.column_number,
            self.injection_order,
            self.batches_label,
        ]
        for col in required_columns:
            if col not in sample_information.columns:
                raise ValueError(f"Column '{col}' not found in sample information.")

        # drop fully blank rows in injection info or columns without names in data
        sample_information = sample_information.dropna(axis="index", how="all")
        has_col_name = ~raw_data.columns.str.startswith("Unnamed: ")
        raw_data = raw_data.loc[:, has_col_name]
        sample_information[self.batches_label] = (
            sample_information[self.batches_label].fillna("").str.replace("_", " ")
        )
        sample_information = sample_information.astype(
            {self.injection_type: str, self.batches_label: str}
        )

        try:
            sample_information = sample_information.astype({self.injection_order: int})
        except ValueError as e:
            raise TypeError(
                "There are missing or non-numeric values in the injection_order."
            ) from e

        if not sample_information[self.injection_order].is_unique:
            raise ValueError("There are duplicate values in the injection_order.")

        if not sample_information[self.injection_id].is_unique:
            raise ValueError("There are duplicate injection_ids.")
        sample_information = sample_information.set_index(self.injection_id, drop=False)

        if not (
            sample_information.sort_values(by=self.injection_order).index
            == sample_information.index
        ).all():
            raise ValueError("Samples must be sorted in the order they were injected.")

        # check all data samples are in info, and get first sample
        in_data = np.isin(sample_information[self.injection_id], raw_data.columns)
        if not (in_data).any():
            raise ValueError(
                "There are no overlapping samples in your sample information and "
                "sample data sheets."
            )
        first_sample = sample_information.loc[in_data, self.injection_id].iloc[0]
        samples = raw_data.loc[:, first_sample:].columns

        non_dc = sample_information[self.injection_type].str.startswith(
            tuple(["not_used", "brpp", "mm", "blank", "other", "tube_blank", "ms2"])
        )

        # flag unused samples
        will_drop = ~in_data & non_dc
        sample_information.loc[will_drop, self.QCRole] = "NA"

        if len(samples) > len(sample_information.loc[~will_drop]):
            raise ValueError(
                "The number of samples in sample information does not match the number "
                "of data columns. Check for excess samples at the end of the data."
            )

        # check there aren't 'used' samples missing from data sheet
        missing = ~np.isin(
            sample_information.loc[~will_drop, self.injection_id], samples
        )
        if missing.any():
            raise ValueError(
                "The following samples are marked as usable but missing from your "
                "data sheet: "
                + ", ".join(
                    sample_information.loc[~will_drop, self.injection_id][missing]
                )
            )

        # check they are in the same order
        same = sample_information.loc[~will_drop, self.injection_id] == samples
        if not same.all():
            bad_sample = sample_information.loc[
                ~will_drop & ~same, self.injection_id
            ].iloc[0]
            raise ValueError(
                f"Your usable samples do not match, starting with '{bad_sample}' "
                f"on your information sheet."
            )

        return raw_data, sample_information, first_sample

    def _parse_data(self, raw_data):
        """
        Separates non-data columns from sample data columns, and cleans the data as
        needed before calculations are performed.
        :param raw_data: DataFrame, raw dataset
        :return:
        """

        raw_data = raw_data.reset_index(drop=True)
        first_sample_idx = raw_data.columns.get_loc(self.first_sample)
        metadata = raw_data.iloc[:, :first_sample_idx]
        data = raw_data.iloc[:, first_sample_idx:]

        try:
            data = data.astype("float64")
        except ValueError as e:
            character = re.search("'.+'", str(e))
            if character:
                raise TypeError(
                    f"Data contains a non-numeric character: " f"{character.group()}."
                ) from None
            raise TypeError("Data contains a non-numeric character.") from e
        data = data.replace(0, np.nan)

        if self.met_col in metadata.columns:
            metadata = metadata.astype({self.met_col: str})
        metadata = metadata.replace("nan", np.nan)

        return data, metadata

    def _find_pools(self, pool_names):
        """
        Locates file names of pooled reference samples.
        :param pool_names: tuple[str], pooled reference sample names
        :return: two params
            dict: pool names as keys, list of file names of pools as values
            pd series: list of not-used pools
        """
        pools = {}
        a, b = pool_names
        info = self.sample_info.sort_values(by=self.injection_order)

        # identify pools based on sample type
        # non-case sensitive, for the time being
        pools[a] = info.loc[
            info[self.injection_type].str.lower() == a.lower(), self.injection_id
        ]
        pools[b] = info.loc[
            info[self.injection_type].str.lower() == b.lower(), self.injection_id
        ]

        # not used -- if it starts with "not_used" and contains a pool name
        not_used_pools = info.loc[
            info[self.injection_type].str.contains(f"{a}|{b}", case=False)
            & info[self.injection_type].str.startswith("not_used"),
            self.injection_id,
        ]

        return pools, not_used_pools

    def _gen_batches(self):
        """
        Splits data into default batches based on column breaks and override batches
        based on user-supplied batch start/end.
        :return:
            default_batches - list, 2D nested containing sample file names for each
            batch
            override_batches - list, 2D nested containing sample file names for each
            batch
        """

        default_batches = []
        override_batches = []
        valid = self.sample_info[self.QCRole] != "NA"

        # default batches split by column number
        for col in pd.unique(self.sample_info[self.column_number]):
            batch = self.sample_info.loc[
                valid & (self.sample_info[self.column_number] == col), self.injection_id
            ]
            if len(batch) > 0:
                default_batches.append(batch)

        # override batches split by user-indicated batch_end
        sample_info = self.sample_info[valid]
        batch_info = sample_info[self.batches_label].str.lower()
        ends = batch_info == "batch_end"
        starts = np.roll(ends, 1)

        for start, end in zip(np.where(starts)[0], np.where(ends)[0]):
            this_batch = sample_info.iloc[start : end + 1, :]
            override_batches.append(this_batch[self.injection_id])

            if (
                sample_info.loc[sample_info.index[start], self.batches_label]
                != "batch_start"
            ):
                warn(
                    "The following sample is being used as a batch start but lacks a "
                    "'batch_start' label: "
                    + sample_info.loc[sample_info.index[start], self.injection_id]
                )

            contains_starts = np.where(this_batch[self.batches_label] == "batch_start")[
                0
            ]
            if (contains_starts > 0).any():
                warn(
                    "The following samples are NOT being used as a batch start but "
                    "have a 'batch_start' label: "
                    + ", ".join(
                        [
                            this_batch.loc[this_batch.index[i], self.injection_id]
                            for i in contains_starts
                            if i != 0
                        ]
                    )
                )

        return default_batches, override_batches

    def internal_standard_correct(self, internal_standards):
        """
        Performs drift correction using one or more internal standards as a reference.
        :param internal_standards: str or array-like, single internal standard or
        list of internal standards to use for drift correction
        """
        if isinstance(internal_standards, str):
            internal_standards = [internal_standards]
        if self.met_col not in self.metadata.columns:
            raise ValueError(
                f"'{self.met_col}' was not found in your columns. Cannot find internal "
                "standard(s) for drift correction."
            )

        is_nonquant = (
            np.array([False] * len(self.metadata))
            if self.Non_Quant not in self.metadata.columns
            else self.metadata[self.Non_Quant].fillna(False)
        )
        if not set(internal_standards).issubset(
            set(self.metadata.loc[~is_nonquant, self.met_col])
        ):
            missing = set(internal_standards) - set(
                self.metadata.loc[~is_nonquant, self.met_col]
            )
            raise ValueError(
                "Internal standard(s) " + str(missing)[1:-1] + " not found in "
                "Metabolite list."
            )

        standards_index = (
            np.isin(self.metadata[self.met_col], internal_standards) & ~is_nonquant
        )
        scalers = self.data.loc[standards_index]
        scalers = scalers.transpose() / scalers.median(axis=1)
        scalers = scalers.mean(axis=1)

        self.cvs.loc[standards_index, "Internal Standards"] = "Internal Standard"

        if scalers.isna().any():
            raise ValueError(
                "The following samples cannot be corrected due to missing values in "
                f"the internal standard(s): {self.data.columns[scalers.isna()].values}"
            )

        self.data = self.data / scalers

        # store drift correction parameters for reporting later
        self.correction_params["IS"] = list(internal_standards)

    def pool_correct(
        self, interpolation, pool, override=True, max_missing_percent=30
    ):  # pylint: disable=too-many-branches
        """
        Performs pool-based drift correction.
        :param interpolation: str, interpolation method: NN or linear
        :param pool: str, pool name to use for correction (typically PREFA or PREFB)
        :param override: boolean, uses override batches in place of default batches
        if true
        :param max_missing_percent: int (0-100), skip drift correction for any row
        that has this percent (or more) drift correction pools missing values
        """

        if pool not in self.pools:
            raise ValueError(
                f"'{pool}' is not a valid pool for this dataset. Valid pools: "
                f"{', '.join(self.pools)}."
            )

        methods = {
            "nn": self._nn_interpolation,
            "linear": self._linear_interpolation,
        }
        self.correction_params[interpolation.lower()] = (pool, override)
        interpolation = methods[interpolation.lower()]
        pool_names = self.pools[pool]
        pool_medians = self.data[pool_names].median(axis=1)
        pool_medians = pool_medians.fillna(1)
        batches = self.batches["override"] if override else self.batches["default"]
        self._set_pool_roles(pool)

        # begin the scaling
        scalers = [[] for _ in self.data.index]
        data = self.data.values

        to_skip = self._flag_skipped_rows(pool, max_missing_percent, len(batches))
        batches_skipped = [[] for _ in self.cvs.index]

        for this_batch in batches:
            batch_pools = this_batch.loc[this_batch.isin(pool_names)]
            batch_pools = pd.DataFrame(batch_pools).join(
                self.sample_info[self.injection_order]
            )
            this_batch = pd.DataFrame(this_batch).join(
                self.sample_info[self.injection_order]
            )
            skipped_str = f"{this_batch.iloc[0,0]} to {this_batch.iloc[-1,0]}"

            if len(batch_pools) == 0:
                raise ValueError(
                    "The following batch does not contain any drift "
                    "correction pools: "
                    f"{this_batch.iloc[0, 0]} to "
                    f"{this_batch.iloc[-1, 0]}"
                )

            batch_scalers = []
            batch_pool_values = data[:, np.isin(self.data.columns, batch_pools.index)]
            batch_pool_order = batch_pools[self.injection_order].values

            for row in tqdm(range(len(data))):
                pool_values = batch_pool_values[row]
                pool_order = batch_pool_order[pool_values > 0]
                pool_values = pool_values[pool_values > 0]

                # skip if previously flagged
                if to_skip[row]:
                    batch_scalers.append([pool_medians[row]] * len(this_batch))

                # skip if there are no pools
                elif len(pool_order) < 1:
                    batch_scalers.append([pool_medians[row]] * len(this_batch))
                    batches_skipped[row].append(skipped_str)

                # "nearest neighbor" if there is 1 pool
                elif len(pool_order) < 2:
                    batch_scalers.append([pool_values[0]] * len(this_batch))

                else:
                    batch_scalers.append(
                        interpolation(
                            this_batch, pool_order, pool_values, self.injection_order
                        )
                    )
            scalers = np.concatenate((scalers, batch_scalers), axis=1)
        scalers = pd.DataFrame(scalers, columns=self.data.columns)
        scalers = scalers.divide(pool_medians, axis="index")
        self.data = self.data.divide(scalers)

        self.cvs[self.Batches_Skipped] = [
            "Not Pool Corrected" if to_skip[i] or len(item) == len(batches) else item
            for i, item in enumerate(batches_skipped)
        ]

    def _set_pool_roles(self, pool):
        dc_pools = set(self.pools[pool].index)
        qc_pools = set()
        for pool_list in self.pools.values():
            qc_pools.update(pool_list.index)
        qc_pools -= dc_pools
        self.sample_info.loc[self.sample_info[self.QCRole] != "NA", self.QCRole] = (
            "Sample"
        )
        self.sample_info.loc[list(dc_pools), self.QCRole] = "QC-drift_correction"
        self.sample_info.loc[list(qc_pools), self.QCRole] = "QC-pooled_ref"
        self.sample_info.loc[self.not_used_pools.index, self.QCRole] = "QC-not_used"

    def _flag_skipped_rows(self, pool, max_missing_percent, batch_count):
        """
        Flag rows that can't be drift corrected: they have zero pools, they have fewer
        than the minimum required pools, or there is only one batch and only one pool
        :param pool: str, drift correction pool (typically PREFA or PREFB)
        :param max_missing_percent: int, max percent of pools allowed to be missing
        :param batch_count: int, number of batches used for pool correction
        :return: list of bools, skip or don't skip for each row of data
        """
        pools = np.isin(self.data.columns, self.pools[pool])
        min_present = (1 - max_missing_percent / 100) * len(self.pools[pool])
        pool_data = self.data.loc[:, pools].values
        pool_counts = np.sum(pool_data > 0, axis=1)
        to_skip = (pool_counts == 0) | (pool_counts < min_present)
        if batch_count == 1:
            to_skip = to_skip | (pool_counts == 1)
        return to_skip

    @staticmethod
    def _nn_interpolation(
        batch, pool_order, pool_values, order_label="injection_order"
    ):
        """
        :param batch: DataFrame, sample file names and injection order for the current
        batch
        :param pool_order: array-like, injection order for each pool in the batch,
        for a single metabolite
        :param pool_values: array-like, pool value for each pool in the batch,
        for a single metabolite
        :return: numpy ndarray, value of nearest neighbor for each sample in the batch
        """

        nn_x = batch[order_label].values
        nn_interp = interpolate.interp1d(
            pool_order,
            pool_values,
            kind="nearest",
            bounds_error=False,
            fill_value=(pool_values[0], pool_values[-1]),
        )
        nn_y = nn_interp(nn_x)

        return nn_y

    @staticmethod
    def _linear_interpolation(
        batch, pool_order, pool_values, order_label="injection_order"
    ):
        """
        :param batch: DataFrame, sample file names and injection order for the current
        batch
        :param pool_order: array-like, injection order for each pool in the batch,
        for a single metabolite
        :param pool_values: array-like, pool value for each pool in the batch,
        for a single metabolite
        :return: numpy ndarray, linearly interpolated value for each sample in the batch
        """

        linear_x = batch[order_label].values
        linear_y = np.interp(x=linear_x, xp=pool_order, fp=pool_values)

        return linear_y

    def calculate_cvs(self):
        """
        Calculates coefficients of variation for each row of data.
        """
        pools = list(self.pools.keys())
        all_data = {  # explicitly define key order, this will be the column order later
            "Raw": self.rdata,
            f"Raw {pools[0]}": self.rdata[self.pools[pools[0]]],
            f"Raw {pools[1]}": self.rdata[self.pools[pools[1]]],
            "Final": self.data,
            pools[0]: self.data[self.pools[pools[0]]],
            pools[1]: self.data[self.pools[pools[1]]],
        }

        for name, data in all_data.items():
            col_name = f"{name} CVs" if name != "Final" else "CVs"
            self.cvs[col_name] = data.std(axis=1) / data.mean(axis=1)

    def to_csv(self, filepath="results.csv", to_string=False):
        """
        Assembles and downloads a csv of the metadata, data, and calculated CVs
        :param filepath: base/beginning of the results filepath
        """
        filepath = filepath.replace(".csv", "")
        for method, params in self.correction_params.items():
            if True in params:
                filepath += f"_{method}_{params[0]}_override"
            elif False in params:
                filepath += f"_{method}_{params[0]}"
            else:
                filepath += f"_{method}"

        # sort CV columns so raw columns come first
        headers = [i for i in self.cvs.columns if "CVs" in i]
        headers.sort(key=lambda x: x.replace("Raw", "AAA"), reverse=True)
        for col in headers:
            self.cvs.insert(0, col, self.cvs.pop(col))

        # make skipped batches lists more readable
        if self.Batches_Skipped in self.cvs.columns:
            self.cvs[self.Batches_Skipped] = self.cvs[self.Batches_Skipped].map(
                lambda x: ", ".join(x) if isinstance(x, list) else x
            )

        output = self.data.apply(np.floor)
        output = output.replace(0, np.nan)

        if self.met_col in self.metadata.columns:
            to_join = [self.cvs, self.metadata.loc[:, self.met_col], output]
        else:
            to_join = [self.cvs, output]
        output = self.metadata.loc[:, self.metadata.columns != self.met_col].join(
            to_join
        )
        if to_string:
            return (
                output.to_csv(lineterminator="\r\n", index=False),
                filepath,
            )
        return output.to_csv(filepath + ".csv", index=False)

    def report(self, filepath="report.csv", to_string=False):
        """
        Writes a report
        :param filepath: report filepath name
        """
        filepath = filepath.replace(".csv", "") + ".csv"
        if to_string:
            return (
                self.sample_info.to_csv(lineterminator="\r\n", index=False),
                filepath,
            )
        return self.sample_info.to_csv(filepath, index=False)


def find_batch_start_end(
    info,
    injection_order_label=IMDATA["injection_order"],
    column_number_label=IMDATA["column_number"],
    batches_label=IMDATA["batches"],
    qc_role_label=IMDATA["QCRole"],
):
    """
    Set the batch start and end labels in the sample information DataFrame. Assumes info
    already has the required columns, is sorted by injection order, and has dropped
    injections flagged as "NA".
    :param info: DataFrame, sample information
    :param injection_order_label: str, column name for injection order
    :param batches_label: str, column name for batch labels
    :param qc_role_label: str, column name for QC role
    :return: DataFrame, updated sample information with batch start and end labels
    """
    info = info.copy()
    # accept "batch_end" and "batch end" as equivalent
    info[batches_label] = info[batches_label].str.replace(" ", "_")
    for item in set(info[batches_label]):
        if item.lower() not in ("batch_start", "batch_end", ""):
            raise ValueError(
                f"An invalid label is present in the batches column: '{item}'. Valid "
                f"batch labels are 'batch_start' and 'batch_end'."
            )

    valid = info[qc_role_label] != "NA"
    # label first valid injection as batch_start
    info.loc[info.loc[valid].index[0], batches_label] = "batch_start"

    # label last valid injection for each column as batch_end
    ends = info.loc[valid, column_number_label].drop_duplicates(keep="last")
    info.loc[ends.index, batches_label] = "batch_end"

    # check that batch_ends aren't on dropped samples
    end_drops = info.loc[~valid, batches_label] == "batch_end"
    if sum(end_drops) > 0:
        end_drop_indices = end_drops.index[end_drops]

        # get valid samples and their injection order
        valid_samples = info.loc[valid]
        valid_orders = valid_samples[injection_order_label]

        for idx in end_drop_indices:
            # find the nearest valid sample with an earlier injection order
            current_order = info.loc[idx, injection_order_label]
            earlier_valid = valid_samples[valid_orders < current_order]
            if earlier_valid.empty:
                raise ValueError(
                    f"Cannot move 'batch_end' label for injection '{idx}' because "
                    "no valid injection exists before it."
                )

            # move the batch_end label to the nearest valid sample
            nearest_valid_idx = earlier_valid.index[-1]
            info.loc[nearest_valid_idx, batches_label] = "batch_end"
            info.loc[idx, batches_label] = ""
        warn(
            "Some 'batch_end' labels were on injections that are not present in the "
            "dataset and have been moved to the nearest valid injections."
        )
    return info[batches_label]
